"""Meta data about a song that USDB shows in the result list."""

from __future__ import annotations

from collections.abc import Iterable
from json import JSONEncoder
from typing import Any, ClassVar

import attrs

from usdb_syncer import SongId, db
from usdb_syncer.constants import UsdbStrings
from usdb_syncer.db import DownloadStatus
from usdb_syncer.sync_meta import SyncMeta


@attrs.define(kw_only=True)
class UsdbSong:
    """Meta data about a song that USDB shows in the result list."""

    song_id: SongId
    artist: str
    title: str
    genre: str
    year: int | None = None
    language: str
    creator: str
    edition: str
    golden_notes: bool
    rating: int
    views: int
    sample_url: str
    # not in USDB song list
    tags: str = ""
    # internal
    sync_meta: SyncMeta | None = None
    status: DownloadStatus = DownloadStatus.NONE
    is_playing: bool = False

    @classmethod
    def from_json(cls, dct: dict[str, Any]) -> UsdbSong:
        dct["song_id"] = SongId(dct["song_id"])
        return cls(**dct)

    @classmethod
    def from_html(
        cls,
        strings: type[UsdbStrings],
        *,
        song_id: str,
        artist: str,
        title: str,
        genre: str,
        year: str,
        language: str,
        creator: str,
        edition: str,
        golden_notes: str,
        rating: str,
        views: str,
        sample_url: str,
    ) -> UsdbSong:
        return cls(
            song_id=SongId.parse(song_id),
            artist=artist,
            title=title,
            genre=genre,
            year=int(year) if len(year) == 4 and year.isdigit() else None,
            language=language,
            creator=creator,
            edition=edition,
            golden_notes=golden_notes == strings.YES,
            rating=rating.count("star.png"),
            views=int(views),
            sample_url=sample_url,
        )

    @classmethod
    def from_db_row(cls, song_id: SongId, row: tuple) -> UsdbSong:
        assert len(row) == 36
        return cls(
            song_id=song_id,
            artist=row[1],
            title=row[2],
            language=row[3],
            edition=row[4],
            golden_notes=bool(row[5]),  # else would be 0/1 instead of False/True
            rating=row[6],
            views=row[7],
            sample_url=row[8],
            year=row[9],
            genre=row[10],
            creator=row[11],
            tags=row[12],
            status=DownloadStatus(row[13]),
            is_playing=bool(row[14]),
            sync_meta=None if row[15] is None else SyncMeta.from_db_row(row[15:]),
        )

    @classmethod
    def get(cls, song_id: SongId) -> UsdbSong | None:
        if song := _UsdbSongCache.get(song_id):
            return song
        if row := db.get_usdb_song(song_id):
            song = UsdbSong.from_db_row(song_id, row)
            _UsdbSongCache.update(song)
            return song
        return None

    def delete(self) -> None:
        db.delete_usdb_song(self.song_id)
        _UsdbSongCache.remove(self.song_id)

    def remove_sync_meta(self) -> None:
        if self.sync_meta:
            self.sync_meta.delete()
            self.sync_meta = None
            _UsdbSongCache.update(self)

    @classmethod
    def delete_all(cls) -> None:
        db.delete_all_usdb_songs()
        _UsdbSongCache.clear()

    def upsert(self) -> None:
        db.upsert_usdb_song(self.db_params())
        db.upsert_usdb_songs_languages([(self.song_id, self.languages())])
        db.upsert_usdb_songs_genres([(self.song_id, self.genres())])
        db.upsert_usdb_songs_creators([(self.song_id, self.creators())])
        if self.sync_meta:
            self.sync_meta.upsert()
        _UsdbSongCache.update(self)

    @classmethod
    def upsert_many(cls, songs: list[UsdbSong]) -> None:
        db.upsert_usdb_songs(song.db_params() for song in songs)
        db.upsert_usdb_songs_languages([(s.song_id, s.languages()) for s in songs])
        db.upsert_usdb_songs_genres([(s.song_id, s.genres()) for s in songs])
        db.upsert_usdb_songs_creators([(s.song_id, s.creators()) for s in songs])
        SyncMeta.upsert_many([song.sync_meta for song in songs if song.sync_meta])
        for song in songs:
            _UsdbSongCache.update(song)

    def db_params(self) -> db.UsdbSongParams:
        return db.UsdbSongParams(
            song_id=self.song_id,
            artist=self.artist,
            title=self.title,
            language=self.language,
            edition=self.edition,
            golden_notes=self.golden_notes,
            rating=self.rating,
            views=self.views,
            sample_url=self.sample_url,
            year=self.year,
            genre=self.genre,
            creator=self.creator,
            tags=self.tags,
            status=self.status,
            is_playing=self.is_playing,
        )

    def is_local(self) -> bool:
        return self.sync_meta is not None and self.sync_meta.path is not None

    def is_pinned(self) -> bool:
        return self.sync_meta is not None and self.sync_meta.pinned

    def languages(self) -> Iterable[str]:
        return (s for lang in self.language.split(",") if (s := lang.strip()))

    def genres(self) -> Iterable[str]:
        return (s for genre in self.genre.split(",") if (s := genre.strip()))

    def creators(self) -> Iterable[str]:
        return (s for creator in self.creator.split(",") if (s := creator.strip()))

    @classmethod
    def clear_cache(cls) -> None:
        _UsdbSongCache.clear()


class UsdbSongEncoder(JSONEncoder):
    """Custom JSON encoder"""

    def default(self, o: Any) -> Any:
        if isinstance(o, UsdbSong):
            fields = attrs.fields(UsdbSong)
            filt = attrs.filters.exclude(
                fields.status, fields.sync_meta, fields.is_playing
            )
            dct = attrs.asdict(o, recurse=False, filter=filt)
            return dct
        return super().default(o)


class _UsdbSongCache:
    """Cache for songs loaded from the DB."""

    _songs: ClassVar[dict[SongId, UsdbSong]] = {}

    @classmethod
    def get(cls, song_id: SongId) -> UsdbSong | None:
        return cls._songs.get(song_id)

    @classmethod
    def update(cls, song: UsdbSong) -> None:
        cls._songs[song.song_id] = song

    @classmethod
    def remove(cls, song_id: SongId) -> None:
        if song_id in cls._songs:
            del cls._songs[song_id]

    @classmethod
    def clear(cls) -> None:
        cls._songs = {}



"""Utilities for running a task in the background while showing a progress dialog."""

import time
import traceback
from collections.abc import Callable
from typing import Any, Generic, TypeVar

import attrs
from PySide6 import QtCore, QtGui, QtWidgets

from usdb_syncer import db, utils
from usdb_syncer.logger import logger

T = TypeVar("T")
_MINIMUM_DURATION_MS = 1000


@attrs.define
class _Error:
    """Trivial wrapper for an exception."""

    error: Exception


@attrs.define
class Result(Generic[T]):
    """The result of an operation, either a value or an error."""

    _result: T | _Error

    def result(self) -> T:
        """If the operation was successful, returns its result, and raises the original
        error otherwise.
        """
        if isinstance(self._result, _Error):
            raise self._result.error
        return self._result

    def log_error(self) -> None:
        """If there was an error, log it without raising."""
        if isinstance(self._result, _Error):
            logger.error(traceback.format_exception(self._result.error))


class _ResultSignal(QtCore.QObject):
    """Signal for when the result is ready."""

    result = QtCore.Signal()


class ProgressDialog(QtWidgets.QProgressDialog):
    """Progress dialog that cannot be closed by the user."""

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:  # noqa: N802
        event.ignore()


def run_with_progress(
    label: str,
    task: Callable[[], T],
    on_done: Callable[[Result[T]], Any] = lambda res: res.result(),
) -> None:
    """Runs a task on a background thread while a modal progress dialog is shown."""
    dialog = ProgressDialog(parent=None, labelText=label, minimum=0, maximum=0)
    dialog.setCancelButton(None)  # type: ignore
    dialog.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)
    dialog.setWindowTitle("USDB Syncer")
    dialog.setMinimumDuration(_MINIMUM_DURATION_MS)
    dialog.setValue(0)
    signal = _ResultSignal()
    result: Result | None = None

    def wrapped_task() -> None:
        nonlocal result
        try:
            with db.managed_connection(utils.AppPaths.db):
                result = Result(task())
        except Exception as exc:  # noqa: BLE001
            result = Result(_Error(exc))
        signal.result.emit()

    def wrapped_on_done() -> None:
        assert result
        dialog.deleteLater()
        on_done(result)
        # prevent Qt from cleaning up the signal before it's done
        # https://bugreports.qt.io/browse/PYSIDE-2921
        _ = signal

    signal.result.connect(wrapped_on_done)
    QtCore.QThreadPool.globalInstance().start(wrapped_task)
    start = time.time()
    while result is None and (time.time() - start) * 1000 < _MINIMUM_DURATION_MS:
        # block until task is completed or dialog shows
        time.sleep(0.01)



"""Remote configuration fetched from GitHub at runtime."""

import functools
import json

import requests

from usdb_syncer import constants
from usdb_syncer.logger import logger
from usdb_syncer.utils import AppPaths

_CONFIG_URL = f"{constants.GITHUB_SHARED_CONTENT}/config.json"
_CONFIG_PATH = AppPaths.shared.joinpath("config.json") if AppPaths.shared else None


@functools.lru_cache(maxsize=1)
def discord_webhook_url() -> str | None:
    if (
        isinstance(dct := _fetch_config().get("discord_webhook"), dict)
        and isinstance(channel := dct.get("channel"), str)
        and isinstance(hash_ := dct.get("hash"), str)
    ):
        return "/".join((constants.DISCORD_WEBHOOK_API, channel, hash_))
    logger.debug("Failed to retrieve Discord webhook URL.")
    return None


@functools.lru_cache(maxsize=1)
def _fetch_config() -> dict:
    if _CONFIG_PATH:
        with _CONFIG_PATH.open(encoding="utf-8") as file:
            return json.load(file)
    return _fetch_remote_config()


def _fetch_remote_config() -> dict:
    try:
        response = requests.get(_CONFIG_URL, timeout=5)
        response.raise_for_status()
        data = response.json()
    except requests.RequestException as error:
        logger.debug(f"Failed to retrieve remote configuration from {_CONFIG_URL}")
        logger.error(str(error))
        data = {}
    else:
        logger.debug(f"Retrieved remote configuration from {_CONFIG_URL}")
    return data



"""Persistent app settings.

To ensure consistent default values and avoid key collisions, QSettings should never be
used directly. Instead, new settings should be added to the SettingKey enum and setters
and getters should be added to this module.
"""

from __future__ import annotations

import os
import shutil
import subprocess
import threading
from enum import Enum, StrEnum, auto
from http.cookiejar import CookieJar
from pathlib import Path
from typing import Any, ClassVar, TypeVar, assert_never, cast

import keyring
import rookiepy
from PySide6.QtCore import QByteArray, QSettings

from usdb_syncer import path_template, utils
from usdb_syncer.constants import Usdb
from usdb_syncer.logger import logger

SYSTEM_USDB = "USDB Syncer/USDB"
NO_KEYRING_BACKEND_WARNING = (
    "Your USDB password cannot be stored or retrieved because no keyring backend is "
    "available. See https://pypi.org/project/keyring for details."
)


def get_usdb_auth() -> tuple[str, str]:
    username = _Settings.get(SettingKey.USDB_USER_NAME, "")
    pwd = ""
    try:
        pwd = keyring.get_password(SYSTEM_USDB, username) or ""
    except keyring.core.backend.errors.NoKeyringError as error:
        logger.debug(error)
        logger.warning(NO_KEYRING_BACKEND_WARNING)
    return (username, pwd)


def set_usdb_auth(username: str, password: str) -> None:
    _Settings.set(SettingKey.USDB_USER_NAME, username)
    try:
        keyring.set_password(SYSTEM_USDB, username, password)
    except keyring.core.backend.errors.NoKeyringError as error:
        logger.debug(error)
        logger.warning(NO_KEYRING_BACKEND_WARNING)


def ffmpeg_is_available() -> bool:
    if shutil.which("ffmpeg"):
        return True
    if (path := get_ffmpeg_dir()) and path not in os.environ["PATH"]:
        # first run; restore path from settings
        utils.add_to_system_path(path)
        if shutil.which("ffmpeg"):
            return True
    return False


class _TemporarySettings:
    """Temporary settings for the USDB Syncer.

    This class is used to store temporary settings in memory.
    It should not be used directly, as it is not thread-safe.
    """

    _temporary_settings: ClassVar[dict[SettingKey, Any]] = {}

    @classmethod
    def remove(cls, key: SettingKey) -> None:
        """Remove a temporary setting."""
        cls._temporary_settings.pop(key, None)

    @classmethod
    def set(cls, key: SettingKey, value: Any) -> None:
        """Set a temporary setting."""
        cls._temporary_settings[key] = value

    @classmethod
    def get(cls, key: SettingKey) -> Any:
        """Get a temporary setting."""
        return cls._temporary_settings.get(key, None)


T = TypeVar("T")


class _Settings:
    """Settings for the USDB Syncer.

    This class is a singleton that provides access to the settings of the USDB Syncer.
    """

    _lock = threading.RLock()

    @classmethod
    def set(cls, key: SettingKey, value: Any, temp: bool = False) -> None:
        if temp:
            cls._set_temporary(key, value)
        else:
            cls._set_permanent(key, value)

    @classmethod
    def _set_temporary(cls, key: SettingKey, value: Any) -> None:
        with cls._lock:
            _TemporarySettings.set(key, value)

    @classmethod
    def _set_permanent(cls, key: SettingKey, value: Any) -> None:
        with cls._lock:
            _TemporarySettings.remove(key)

            if isinstance(value, bool):
                # Qt stores bools as "true" and "false" otherwise
                value = int(value)
            QSettings().setValue(key.value, value)

    @classmethod
    def get(cls, key: SettingKey, default: T) -> T:
        with cls._lock:
            if temp := _TemporarySettings.get(key):
                return temp

            try:
                value = QSettings().value(key.value)
            except (AttributeError, ValueError):
                # setting contains a type incompatible with this version
                return default
            if isinstance(value, ret_type := type(default)):
                return value
            if isinstance(default, bool) and isinstance(value, int):
                # we store bools as ints because Qt doesn't store raw bools
                return cast(T, bool(value))
            if isinstance(value, str) and ret_type in (int, float, bool):
                # in INI files (default on Linux) numeric values are stored as strings
                try:
                    if ret_type is bool:
                        value = int(value)
                    # MyPy doesn't understand any of this, but it should be safe
                    return cast(T, ret_type(value))  # type: ignore[call-arg]
                except (ValueError, TypeError):
                    pass
            return default

    @classmethod
    def reset(cls) -> None:
        """Reset QSettings to default."""
        QSettings().clear()


class SettingKey(Enum):
    """Keys for storing and retrieving settings."""

    SONG_DIR = "song_dir"
    FFMPEG_DIR = "ffmpeg_dir"
    BROWSER = "downloads/browser"
    TXT = "downloads/txt"
    ENCODING = "downloads/encoding"
    NEWLINE = "downloads/newline"
    FORMAT_VERSION = "downloads/format_version"
    FIX_LINEBREAKS = "fixes/linebreaks"
    FIX_FIRST_WORDS_CAPITALIZATION = "fixes/firstwordscapitalization"
    FIX_SPACES = "fixes/spaces"
    FIX_QUOTATION_MARKS = "fixes/quotation_marks"
    THROTTLING_THREADS = "downloads/throttling_threads"
    YTDLP_RATE_LIMIT = "downloads/ytdlp_rate_limit"
    AUDIO = "downloads/audio"
    AUDIO_FORMAT = "downloads/audio_format"
    AUDIO_BITRATE = "downloads/audio_bitrate"
    AUDIO_NORMALIZATION = "downloads/audio_normalization"
    AUDIO_EMBED_ARTWORK = "downloads/audio_embed_artwork"
    VIDEO = "downloads/video"
    VIDEO_FORMAT = "downloads/video_format"
    VIDEO_REENCODE = "downloads/video_reencode"
    VIDEO_FORMAT_NEW = "downloads/video_format_new"
    VIDEO_RESOLUTION_MAX = "downloads/video_resolution_max"
    VIDEO_FPS_MAX = "downloads/video_fps_max"
    VIDEO_EMBED_ARTWORK = "downloads/video_embed_artwork"
    COVER = "downloads/cover"
    COVER_MAX_SIZE = "downloads/cover_max_size"
    BACKGROUND = "downloads/background"
    BACKGROUND_ALWAYS = "downloads/background_always"
    DISCORD_ALLOWED = "downloads/discord_allowed"
    MAIN_WINDOW_GEOMETRY = "geometry/main_window"
    DOCK_LOG_GEOMETRY = "geometry/dock_log"
    MAIN_WINDOW_STATE = "state/main_window"
    TABLE_VIEW_HEADER_STATE = "list_view/header/state"
    USDB_USER_NAME = "usdb/username"
    PATH_TEMPLATE = "files/path_template"
    APP_PATH_KAREDI = "app_paths/karedi"
    APP_PATH_PERFORMOUS = "app_paths/performous"
    APP_PATH_ULTRASTAR_MANAGER = "app_paths/ultrastar_manager"
    APP_PATH_USDX = "app_paths/usdx"
    APP_PATH_VOCALUXE = "app_paths/vocaluxe"
    APP_PATH_YASS_RELOADED = "app_paths/yass_reloaded"
    REPORT_PDF_PAGESIZE = "report/pdf_pagesize"
    REPORT_PDF_ORIENTATION = "report/pdf_orientation"
    REPORT_PDF_MARGIN = "report/pdf_margin"
    REPORT_PDF_COLUMNS = "report/pdf_columns"
    REPORT_PDF_FONTSIZE = "report/pdf_fontsize"
    REPORT_JSON_INDENT = "report/json_indent"
    VIEW_THEME = "view/theme"
    VIEW_PRIMARY_COLOR = "view/primary_color"
    VIEW_COLORED_BACKGROUND = "view/colored_background"


class Encoding(Enum):
    """Supported encodings for song txts."""

    UTF_8 = "utf_8"
    UTF_8_BOM = "utf_8_sig"
    CP1252 = "cp1252"

    def __str__(self) -> str:
        match self:
            case Encoding.UTF_8:
                return "UTF-8"
            case Encoding.UTF_8_BOM:
                return "UTF-8 BOM (legacy support for older Vocaluxe versions)"
            case Encoding.CP1252:
                return "CP1252 (legacy support for older USDX CMD)"
            case _ as unreachable:
                assert_never(unreachable)


class Newline(Enum):
    """Supported line endings for song txts."""

    LF = "\n"
    CRLF = "\r\n"

    def __str__(self) -> str:
        match self:
            case Newline.LF:
                return "Mac/Linux (LF)"
            case Newline.CRLF:
                return "Windows (CRLF)"
            case _ as unreachable:
                assert_never(unreachable)

    @staticmethod
    def default() -> Newline:
        if os.linesep == Newline.CRLF.value:
            return Newline.CRLF
        return Newline.LF


class FormatVersion(Enum):
    """Supported format versions for song txts."""

    V1_0_0 = "1.0.0"
    V1_1_0 = "1.1.0"
    V1_2_0 = "1.2.0"

    def __str__(self) -> str:
        return str(self.value)


class FixLinebreaks(Enum):
    """Supported variants for fixing linebreak timings."""

    DISABLE = 0
    USDX_STYLE = 1
    YASS_STYLE = 2

    def __str__(self) -> str:
        match self:
            case FixLinebreaks.DISABLE:
                return "disable"
            case FixLinebreaks.USDX_STYLE:
                return "USDX style"
            case FixLinebreaks.YASS_STYLE:
                return "YASS style"
            case _ as unreachable:
                assert_never(unreachable)


class FixSpaces(Enum):
    """Supported variants for fixing spaces."""

    DISABLE = 0
    AFTER = 1
    BEFORE = 2

    def __str__(self) -> str:
        match self:
            case FixSpaces.DISABLE:
                return "disable"
            case FixSpaces.AFTER:
                return "after words"
            case FixSpaces.BEFORE:
                return "before words"
            case _ as unreachable:
                assert_never(unreachable)


class CoverMaxSize(Enum):
    """Maximum cover size."""

    DISABLE = 0
    PX_1920 = 1920
    PX_1000 = 1000
    PX_640 = 640

    def __str__(self) -> str:
        match self:
            case CoverMaxSize.DISABLE:
                return "disable"
            case CoverMaxSize.PX_1920:
                return "1920x1920 px"
            case CoverMaxSize.PX_1000:
                return "1000x1000 px"
            case CoverMaxSize.PX_640:
                return "640x640 px"
            case _ as unreachable:
                assert_never(unreachable)


class YtdlpRateLimit(Enum):
    """Rate limits for yt-dlp (B/s)."""

    DISABLE = None
    KIBS_500 = 500 * 1024
    KIBS_1000 = 1000 * 1024
    KIBS_2000 = 2000 * 1024
    KIBS_3000 = 3000 * 1024
    KIBS_4000 = 4000 * 1024

    def __str__(self) -> str:
        if self.value is not None:
            return f"{self.value // 1024} KiB/s"
        return "disabled"


class AudioFormat(Enum):
    """Audio containers that can be requested when downloading with ytdl."""

    M4A = "m4a"
    MP3 = "mp3"
    OGG = "ogg"
    OPUS = "opus"

    def __str__(self) -> str:
        match self:
            case AudioFormat.M4A:
                return ".m4a (mp4a)"
            case AudioFormat.MP3:
                return ".mp3 (MPEG)"
            case AudioFormat.OGG:
                return ".ogg (Ogg Vorbis)"
            case AudioFormat.OPUS:
                return ".opus (Ogg Opus)"
            case _ as unreachable:
                assert_never(unreachable)

    def ytdl_format(self) -> str:
        # prefer best audio-only codec; otherwise take a video codec and extract later
        return f"bestaudio[ext={self.value}]/bestaudio/bestaudio*"

    def ytdl_codec(self) -> str:
        match self:
            case AudioFormat.M4A:
                return "m4a"
            case AudioFormat.MP3:
                return "mp3"
            case AudioFormat.OGG:
                return "vorbis"
            case AudioFormat.OPUS:
                return "opus"
            case _ as unreachable:
                assert_never(unreachable)

    def ffmpeg_encoder(self) -> str:
        match self:
            case AudioFormat.M4A:
                return "aac"
            case AudioFormat.MP3:
                return "libmp3lame"
            case AudioFormat.OGG:
                return "libvorbis"
            case AudioFormat.OPUS:
                return "libopus"
            case _ as unreachable:
                assert_never(unreachable)


class AudioBitrate(Enum):
    """Audio bitrate."""

    KBPS_128 = "128 kbps"
    KBPS_160 = "160 kbps"
    KBPS_192 = "192 kbps"
    KBPS_256 = "256 kbps"
    KBPS_320 = "320 kbps"

    def __str__(self) -> str:
        return self.value

    def ytdl_format(self) -> str:
        return self.value.removesuffix(" kbps")

    def ffmpeg_format(self) -> int:
        return int(self.value.removesuffix(" kbps")) * 1000  # in bits/s


class AudioNormalization(Enum):
    """Audio normalization."""

    DISABLE = None
    REPLAYGAIN = "ReplayGain"
    NORMALIZE = "Normalize (rewrites file)"

    def __str__(self) -> str:
        if self.value is not None:
            return self.value
        return "disabled"


class Browser(Enum):
    """Browsers to use cookies from."""

    NONE = None
    ARC = "arc"
    BRAVE = "brave"
    CHROME = "chrome"
    CHROMIUM = "chromium"
    EDGE = "edge"
    FIREFOX = "firefox"
    LIBREWOLF = "librewolf"
    OCTO_BROWSER = "octo browser"
    OPERA = "opera"
    OPERA_GX = "opera gx"
    SAFARI = "safari"
    VIVALDI = "vivaldi"

    def __str__(self) -> str:
        if self is Browser.NONE:
            return "None"
        return self.value.capitalize()

    def cookies(self) -> CookieJar | None:  # noqa: C901
        match self:
            case Browser.NONE:
                return None
            case Browser.ARC:
                function = rookiepy.arc
            case Browser.BRAVE:
                function = rookiepy.brave
            case Browser.CHROME:
                function = rookiepy.chrome
            case Browser.CHROMIUM:
                function = rookiepy.chromium
            case Browser.EDGE:
                function = rookiepy.edge
            case Browser.FIREFOX:
                function = rookiepy.firefox
            case Browser.LIBREWOLF:
                function = rookiepy.librewolf
            case Browser.OCTO_BROWSER:
                function = rookiepy.octo_browser
            case Browser.OPERA:
                function = rookiepy.opera
            case Browser.OPERA_GX:
                function = rookiepy.opera_gx
            case Browser.SAFARI:
                function = rookiepy.safari
            case Browser.VIVALDI:
                function = rookiepy.vivaldi
            case _ as unreachable:
                assert_never(unreachable)
        try:
            return rookiepy.to_cookiejar(function([Usdb.DOMAIN]))
        except Exception:  # noqa: BLE001
            logger.exception(None)
        logger.warning(f"Failed to retrieve {self!s} cookies.")
        return None


class VideoContainer(Enum):
    """Video containers that can be requested when downloading with ytdl."""

    BEST = "bestvideo"
    MP4 = "mp4"
    MP4_AV1 = "mp4_av1"
    MP4_AVC = "mp4_avc"
    MP4_VP9 = "mp4_vp9"
    WEBM = "webm"

    def __str__(self) -> str:
        match self:
            case VideoContainer.BEST:
                return "Best available container/codec"
            case VideoContainer.MP4:
                return ".mp4 (best available codec)"
            case VideoContainer.MP4_AV1:
                return ".mp4 (AV1 codec)"
            case VideoContainer.MP4_AVC:
                return ".mp4 (AVC/H.264 codec)"
            case VideoContainer.MP4_VP9:
                return ".mp4 (VP9 codec)"
            case VideoContainer.WEBM:
                return ".webm (VP9 codec)"
            case _ as unreachable:
                assert_never(unreachable)

    def ytdl_ext(self) -> str | None:
        match self:
            case VideoContainer.BEST:
                return None
            case (
                VideoContainer.MP4
                | VideoContainer.MP4_AV1
                | VideoContainer.MP4_AVC
                | VideoContainer.MP4_VP9
            ):
                return VideoContainer.MP4.value
            case VideoContainer.WEBM:
                return self.value
            case _ as unreachable:
                assert_never(unreachable)

    def ytdl_vcodec(self) -> str | None:
        match self:
            case VideoContainer.BEST:
                return None
            case VideoContainer.MP4 | VideoContainer.WEBM:
                return None
            case VideoContainer.MP4_AV1:
                return "av01"
            case VideoContainer.MP4_AVC:
                return "^(avc|h264)"
            case VideoContainer.MP4_VP9:
                return "^vp0?9"
            case _ as unreachable:
                assert_never(unreachable)


class VideoCodec(Enum):
    """Video codecs that ytdl can reencode videos to."""

    H264 = "h264"
    H265 = "h265"
    LIBVPX = "libvpx-vp9"
    LIBAOM = "libaom-av1"

    def __str__(self) -> str:
        match self:
            case VideoCodec.H264:
                return "h264"
            case VideoCodec.H265:
                return "h265"
            case VideoCodec.LIBVPX:
                return "libvpx-vp9"
            case VideoCodec.LIBAOM:
                return "libaom-av1"
            case _ as unreachable:
                assert_never(unreachable)


class VideoResolution(Enum):
    """Maximum video resolution."""

    P2160 = "2160p"
    P1440 = "1440p"
    P1080 = "1080p"
    P720 = "720p"
    P480 = "480p"
    P360 = "360p"

    def __str__(self) -> str:
        return self.value

    def width(self) -> int:
        match self:
            case VideoResolution.P2160:
                return 3840
            case VideoResolution.P1440:
                return 2560
            case VideoResolution.P1080:
                return 1920
            case VideoResolution.P720:
                return 1280
            case VideoResolution.P480:
                return 854
            case VideoResolution.P360:
                return 640
            case _ as unreachable:
                assert_never(unreachable)

    def height(self) -> int:
        match self:
            case VideoResolution.P2160:
                return 2160
            case VideoResolution.P1440:
                return 1440
            case VideoResolution.P1080:
                return 1080
            case VideoResolution.P720:
                return 720
            case VideoResolution.P480:
                return 480
            case VideoResolution.P360:
                return 360
            case _ as unreachable:
                assert_never(unreachable)


class VideoFps(Enum):
    """Maximum frames per second."""

    FPS_60 = 60
    FPS_30 = 30

    def __str__(self) -> str:
        return str(self.value)


class Theme(Enum):
    """Application theme."""

    SYSTEM = "System"
    DARK = "Dark"

    def __str__(self) -> str:
        return self.value


class Color(Enum):
    """Colors for GUI customization."""

    RED = "Red"
    PINK = "Pink"
    PURPLE = "Purple"
    DEEPPURPLE = "Deep purple"
    INDIGO = "Indigo"
    BLUE = "Blue"
    LIGHTBLUE = "Light blue"
    CYAN = "Cyan"
    TEAL = "Teal"
    GREEN = "Green"
    LIGHTGREEN = "Light green"
    LIME = "Lime"
    YELLOW = "Yellow"
    AMBER = "Amber"
    ORANGE = "Orange"
    DEEPORANGE = "Deep orange"
    BROWN = "Brown"
    GRAY = "Gray"
    BLUEGRAY = "Blue gray"

    def __str__(self) -> str:
        return self.value


class SupportedApps(StrEnum):
    """Supported third-party apps to be launched from the USDB Syncer."""

    KAREDI = auto()
    PERFORMOUS = auto()
    ULTRASTAR_MANAGER = auto()
    USDX = auto()
    VOCALUXE = auto()
    YASS_RELOADED = auto()

    def __str__(self) -> str:
        match self:
            case SupportedApps.KAREDI:
                return "Karedi"
            case SupportedApps.PERFORMOUS:
                return "Performous"
            case SupportedApps.ULTRASTAR_MANAGER:
                return "UltraStar Manager"
            case SupportedApps.USDX:
                return "UltraStar Deluxe"
            case SupportedApps.VOCALUXE:
                return "Vocaluxe"
            case SupportedApps.YASS_RELOADED:
                return "YASS Reloaded"
            case _ as unreachable:
                assert_never(unreachable)

    def executable_name(self) -> str:
        match self:
            case SupportedApps.KAREDI:
                return "Karedi"
            case SupportedApps.PERFORMOUS:
                return "performous"
            case SupportedApps.ULTRASTAR_MANAGER:
                return "UltraStar-Manager"
            case SupportedApps.USDX:
                return "ultrastardx"
            case SupportedApps.VOCALUXE:
                return "Vocaluxe"
            case SupportedApps.YASS_RELOADED:
                return "yass"
            case _ as unreachable:
                assert_never(unreachable)

    def songpath_parameter(self) -> str:
        match self:
            case SupportedApps.KAREDI:
                return ""
            case SupportedApps.PERFORMOUS:
                return ""
            case SupportedApps.ULTRASTAR_MANAGER:
                return "-songpath"
            case SupportedApps.USDX:
                return "-songpath"
            case SupportedApps.VOCALUXE:
                return "-SongFolder"
            case SupportedApps.YASS_RELOADED:
                return ""
            case _ as unreachable:
                assert_never(unreachable)

    def open_app(self, path: Path) -> None:
        logger.debug(f"Starting {self} with '{path}'.")
        executable = get_app_path(self)
        if executable is None:
            return
        if executable.suffix == ".jar":
            cmd = ["java", "-jar", str(executable), str(path)]
        else:
            cmd = [str(executable), self.songpath_parameter(), str(path)]
        try:
            utils.start_process_detached(cmd)
        except FileNotFoundError:
            logger.error(
                f"Failed to launch {self} from '{executable!s}', file not found. "
                "Please check the executable path in the settings."
            )
        except OSError:
            logger.exception(
                f"Failed to launch {self} from '{executable!s}', I/O error."
            )
        except subprocess.SubprocessError:
            logger.exception(
                f"Failed to launch {self} from '{executable!s}', subprocess error."
            )


class ReportPDFPagesize(Enum):
    """Supported PDF page sizes."""

    A3 = "A3"
    A4 = "A4"
    A5 = "A5"
    LETTER = "Letter"
    LEGAL = "Legal"

    def __str__(self) -> str:
        return str(self.value)


class ReportPDFOrientation(Enum):
    """Supported PDF page orientations."""

    PORTRAIT = "Portrait"
    LANDSCAPE = "Landscape"

    def __str__(self) -> str:
        return str(self.value)


def reset() -> None:
    _Settings.reset()


def get_throttling_threads() -> int:
    return _Settings.get(SettingKey.THROTTLING_THREADS, 0)


def set_throttling_threads(value: int, temp: bool = False) -> None:
    _Settings.set(SettingKey.THROTTLING_THREADS, value, temp)


def get_ytdlp_rate_limit() -> YtdlpRateLimit:
    return _Settings.get(SettingKey.YTDLP_RATE_LIMIT, YtdlpRateLimit.DISABLE)


def set_ytdlp_rate_limit(value: YtdlpRateLimit, temp: bool = False) -> None:
    _Settings.set(SettingKey.YTDLP_RATE_LIMIT, value, temp)


def get_audio() -> bool:
    return _Settings.get(SettingKey.AUDIO, True)


def set_audio(value: bool, temp: bool = False) -> None:
    _Settings.set(SettingKey.AUDIO, value, temp)


def get_audio_format() -> AudioFormat:
    return _Settings.get(SettingKey.AUDIO_FORMAT, AudioFormat.M4A)


def set_audio_format(value: AudioFormat, temp: bool = False) -> None:
    _Settings.set(SettingKey.AUDIO_FORMAT, value, temp)


def get_audio_bitrate() -> AudioBitrate:
    return _Settings.get(SettingKey.AUDIO_BITRATE, AudioBitrate.KBPS_256)


def set_audio_bitrate(value: AudioBitrate, temp: bool = False) -> None:
    _Settings.set(SettingKey.AUDIO_BITRATE, value, temp)


def get_audio_normalization() -> AudioNormalization:
    return _Settings.get(SettingKey.AUDIO_NORMALIZATION, AudioNormalization.DISABLE)


def set_audio_normalization(value: AudioNormalization, temp: bool = False) -> None:
    _Settings.set(SettingKey.AUDIO_NORMALIZATION, value, temp)


def get_audio_embed_artwork() -> bool:
    return _Settings.get(SettingKey.AUDIO_EMBED_ARTWORK, False)


def set_audio_embed_artwork(value: bool, temp: bool = False) -> None:
    _Settings.set(SettingKey.AUDIO_EMBED_ARTWORK, value, temp)


def get_encoding() -> Encoding:
    return _Settings.get(SettingKey.ENCODING, Encoding.UTF_8)


def set_encoding(value: Encoding, temp: bool = False) -> None:
    _Settings.set(SettingKey.ENCODING, value, temp)


def get_newline() -> Newline:
    return _Settings.get(SettingKey.NEWLINE, Newline.default())


def set_newline(value: Newline, temp: bool = False) -> None:
    _Settings.set(SettingKey.NEWLINE, value, temp)


def get_version() -> FormatVersion:
    return _Settings.get(SettingKey.FORMAT_VERSION, FormatVersion.V1_0_0)


def set_version(value: FormatVersion, temp: bool = False) -> None:
    _Settings.set(SettingKey.FORMAT_VERSION, value, temp)


def get_txt() -> bool:
    return _Settings.get(SettingKey.TXT, True)


def set_txt(value: bool, temp: bool = False) -> None:
    _Settings.set(SettingKey.TXT, value, temp)


def get_fix_linebreaks() -> FixLinebreaks:
    return _Settings.get(SettingKey.FIX_LINEBREAKS, FixLinebreaks.YASS_STYLE)


def set_fix_linebreaks(value: FixLinebreaks, temp: bool = False) -> None:
    _Settings.set(SettingKey.FIX_LINEBREAKS, value, temp)


def get_fix_first_words_capitalization() -> bool:
    return _Settings.get(SettingKey.FIX_FIRST_WORDS_CAPITALIZATION, True)


def set_fix_first_words_capitalization(value: bool, temp: bool = False) -> None:
    _Settings.set(SettingKey.FIX_FIRST_WORDS_CAPITALIZATION, value, temp)


def get_fix_spaces() -> FixSpaces:
    return _Settings.get(SettingKey.FIX_SPACES, FixSpaces.AFTER)


def set_fix_spaces(value: FixSpaces, temp: bool = False) -> None:
    _Settings.set(SettingKey.FIX_SPACES, value, temp)


def get_fix_quotation_marks() -> bool:
    return _Settings.get(SettingKey.FIX_QUOTATION_MARKS, True)


def set_fix_quotation_marks(value: bool, temp: bool = False) -> None:
    _Settings.set(SettingKey.FIX_QUOTATION_MARKS, value, temp)


def get_cover() -> bool:
    return _Settings.get(SettingKey.COVER, True)


def set_cover(value: bool, temp: bool = False) -> None:
    _Settings.set(SettingKey.COVER, value, temp)


def get_cover_max_size() -> CoverMaxSize:
    return _Settings.get(SettingKey.COVER_MAX_SIZE, CoverMaxSize.PX_1920)


def set_cover_max_size(value: CoverMaxSize, temp: bool = False) -> None:
    _Settings.set(SettingKey.COVER_MAX_SIZE, value, temp)


def get_browser() -> Browser:
    return _Settings.get(SettingKey.BROWSER, Browser.CHROME)


def set_browser(value: Browser, temp: bool = False) -> None:
    _Settings.set(SettingKey.BROWSER, value, temp)


def get_song_dir() -> Path:
    return _Settings.get(SettingKey.SONG_DIR, Path("songs").resolve())


def set_song_dir(value: Path, temp: bool = False) -> None:
    _Settings.set(SettingKey.SONG_DIR, value, temp)


def get_video() -> bool:
    return _Settings.get(SettingKey.VIDEO, True)


def set_video(value: bool, temp: bool = False) -> None:
    _Settings.set(SettingKey.VIDEO, value, temp)


def get_video_format() -> VideoContainer:
    return _Settings.get(SettingKey.VIDEO_FORMAT, VideoContainer.MP4)


def set_video_format(value: VideoContainer, temp: bool = False) -> None:
    _Settings.set(SettingKey.VIDEO_FORMAT, value, temp)


def get_video_reencode() -> bool:
    return _Settings.get(SettingKey.VIDEO_REENCODE, False)


def set_video_reencode(value: bool, temp: bool = False) -> None:
    _Settings.set(SettingKey.VIDEO_REENCODE, value, temp)


def get_video_format_new() -> VideoCodec:
    return _Settings.get(SettingKey.VIDEO_FORMAT_NEW, VideoCodec.H264)


def set_video_format_new(value: VideoCodec, temp: bool = False) -> None:
    _Settings.set(SettingKey.VIDEO_FORMAT_NEW, value, temp)


def get_video_resolution() -> VideoResolution:
    return _Settings.get(SettingKey.VIDEO_RESOLUTION_MAX, VideoResolution.P1080)


def set_video_resolution(value: VideoResolution, temp: bool = False) -> None:
    _Settings.set(SettingKey.VIDEO_RESOLUTION_MAX, value, temp)


def get_video_fps() -> VideoFps:
    return _Settings.get(SettingKey.VIDEO_FPS_MAX, VideoFps.FPS_60)


def set_video_fps(value: VideoFps, temp: bool = False) -> None:
    _Settings.set(SettingKey.VIDEO_FPS_MAX, value, temp)


def get_video_embed_artwork() -> bool:
    return _Settings.get(SettingKey.VIDEO_EMBED_ARTWORK, False)


def set_video_embed_artwork(value: bool, temp: bool = False) -> None:
    _Settings.set(SettingKey.VIDEO_EMBED_ARTWORK, value, temp)


def get_background() -> bool:
    return _Settings.get(SettingKey.BACKGROUND, True)


def set_background(value: bool, temp: bool = False) -> None:
    _Settings.set(SettingKey.BACKGROUND, value, temp)


def get_background_always() -> bool:
    return _Settings.get(SettingKey.BACKGROUND_ALWAYS, True)


def set_background_always(value: bool, temp: bool = False) -> None:
    _Settings.set(SettingKey.BACKGROUND_ALWAYS, value, temp)


def get_discord_allowed() -> bool:
    return _Settings.get(SettingKey.DISCORD_ALLOWED, False)


def set_discord_allowed(value: bool, temp: bool = False) -> None:
    _Settings.set(SettingKey.DISCORD_ALLOWED, value, temp)


def get_ffmpeg_dir() -> str:
    return _Settings.get(SettingKey.FFMPEG_DIR, "")


def set_ffmpeg_dir(value: str, temp: bool = False) -> None:
    _Settings.set(SettingKey.FFMPEG_DIR, value, temp)


def get_geometry_main_window() -> QByteArray:
    return _Settings.get(SettingKey.MAIN_WINDOW_GEOMETRY, QByteArray())


def set_geometry_main_window(geometry: QByteArray, temp: bool = False) -> None:
    _Settings.set(SettingKey.MAIN_WINDOW_GEOMETRY, geometry, temp)


def get_state_main_window() -> QByteArray:
    return _Settings.get(SettingKey.MAIN_WINDOW_STATE, QByteArray())


def set_state_main_window(state: QByteArray, temp: bool = False) -> None:
    _Settings.set(SettingKey.MAIN_WINDOW_STATE, state, temp)


def get_geometry_log_dock() -> QByteArray:
    return _Settings.get(SettingKey.DOCK_LOG_GEOMETRY, QByteArray())


def set_geometry_log_dock(state: QByteArray, temp: bool = False) -> None:
    _Settings.set(SettingKey.DOCK_LOG_GEOMETRY, state, temp)


def get_table_view_header_state() -> QByteArray:
    return _Settings.get(SettingKey.TABLE_VIEW_HEADER_STATE, QByteArray())


def set_table_view_header_state(state: QByteArray, temp: bool = False) -> None:
    _Settings.set(SettingKey.TABLE_VIEW_HEADER_STATE, state, temp)


def get_path_template() -> path_template.PathTemplate:
    return _Settings.get(SettingKey.PATH_TEMPLATE, path_template.PathTemplate.default())


def set_path_template(template: path_template.PathTemplate, temp: bool = False) -> None:
    _Settings.set(SettingKey.PATH_TEMPLATE, template, temp)


def get_theme() -> Theme:
    return _Settings.get(SettingKey.VIEW_THEME, Theme.SYSTEM)


def set_theme(theme: Theme) -> None:
    _Settings.set(SettingKey.VIEW_THEME, theme)


def get_primary_color() -> Color:
    return _Settings.get(SettingKey.VIEW_PRIMARY_COLOR, Color.RED)


def set_primary_color(primary_color: Color) -> None:
    _Settings.set(SettingKey.VIEW_PRIMARY_COLOR, primary_color)


def get_colored_background() -> bool:
    return _Settings.get(SettingKey.VIEW_COLORED_BACKGROUND, False)


def set_colored_background(colored_background: bool) -> None:
    _Settings.set(SettingKey.VIEW_COLORED_BACKGROUND, colored_background)


def get_app_path(app: SupportedApps) -> Path | None:
    match app:
        case SupportedApps.KAREDI:
            path = _Settings.get(SettingKey.APP_PATH_KAREDI, "")
        case SupportedApps.PERFORMOUS:
            path = _Settings.get(SettingKey.APP_PATH_PERFORMOUS, "")
        case SupportedApps.ULTRASTAR_MANAGER:
            path = _Settings.get(SettingKey.APP_PATH_ULTRASTAR_MANAGER, "")
        case SupportedApps.USDX:
            path = _Settings.get(SettingKey.APP_PATH_USDX, "")
        case SupportedApps.VOCALUXE:
            path = _Settings.get(SettingKey.APP_PATH_VOCALUXE, "")
        case SupportedApps.YASS_RELOADED:
            path = _Settings.get(SettingKey.APP_PATH_YASS_RELOADED, "")
        case _ as unreachable:
            assert_never(unreachable)
    return Path(path) if path != "" else None


def set_app_path(app: SupportedApps, path: str, temp: bool = False) -> None:
    setting_key = None
    match app:
        case SupportedApps.KAREDI:
            setting_key = SettingKey.APP_PATH_KAREDI
        case SupportedApps.PERFORMOUS:
            setting_key = SettingKey.APP_PATH_PERFORMOUS
        case SupportedApps.ULTRASTAR_MANAGER:
            setting_key = SettingKey.APP_PATH_ULTRASTAR_MANAGER
        case SupportedApps.USDX:
            setting_key = SettingKey.APP_PATH_USDX
        case SupportedApps.VOCALUXE:
            setting_key = SettingKey.APP_PATH_VOCALUXE
        case SupportedApps.YASS_RELOADED:
            setting_key = SettingKey.APP_PATH_YASS_RELOADED
        case _ as unreachable:
            assert_never(unreachable)

    _Settings.set(setting_key, path, temp)


def get_report_pdf_pagesize() -> ReportPDFPagesize:
    return _Settings.get(SettingKey.REPORT_PDF_PAGESIZE, ReportPDFPagesize.A4)


def set_report_pdf_pagesize(pagesize: ReportPDFPagesize, temp: bool = False) -> None:
    _Settings.set(SettingKey.REPORT_PDF_PAGESIZE, pagesize, temp)


def get_report_pdf_orientation() -> ReportPDFOrientation:
    return _Settings.get(
        SettingKey.REPORT_PDF_ORIENTATION, ReportPDFOrientation.PORTRAIT
    )


def set_report_pdf_orientation(
    orientation: ReportPDFOrientation, temp: bool = False
) -> None:
    _Settings.set(SettingKey.REPORT_PDF_ORIENTATION, orientation, temp)


def get_report_pdf_margin() -> int:
    return _Settings.get(SettingKey.REPORT_PDF_MARGIN, 20)


def set_report_pdf_margin(margin: int, temp: bool = False) -> None:
    _Settings.set(SettingKey.REPORT_PDF_MARGIN, margin, temp)


def get_report_pdf_columns() -> int:
    return _Settings.get(SettingKey.REPORT_PDF_COLUMNS, 2)


def set_report_pdf_columns(columns: int, temp: bool = False) -> None:
    _Settings.set(SettingKey.REPORT_PDF_COLUMNS, columns, temp)


def get_report_pdf_fontsize() -> int:
    return _Settings.get(SettingKey.REPORT_PDF_FONTSIZE, 10)


def set_report_pdf_fontsize(fontsize: int, temp: bool = False) -> None:
    _Settings.set(SettingKey.REPORT_PDF_FONTSIZE, fontsize, temp)


def get_report_json_indent() -> int:
    return _Settings.get(SettingKey.REPORT_JSON_INDENT, 4)


def set_report_json_indent(indent: int, temp: bool = False) -> None:
    _Settings.set(SettingKey.REPORT_JSON_INDENT, indent, temp)



"""Import and export lists of USDB IDs from file system."""

from __future__ import annotations

import configparser
import json
from collections.abc import Iterable
from pathlib import Path
from urllib.parse import parse_qs, urlparse

import attrs
import bs4
from bs4 import BeautifulSoup

from usdb_syncer import SongId, errors
from usdb_syncer.logger import logger
from usdb_syncer.usdb_song import UsdbSong


@attrs.define
class UsdbIdFileError(errors.UsdbSyncerError):
    """USDB File Parser root exception"""


@attrs.define
class UsdbIdFileUnsupportedExtensionError(UsdbIdFileError):
    """file extension is not supported for parsing"""

    def __str__(self) -> str:
        return "file extension is not supported"


@attrs.define
class UnexpectedUsdbIdFileError(UsdbIdFileError):
    """Unknown cause while reading file"""

    def __str__(self) -> str:
        return "Unexpected error reading file"


@attrs.define
class UsdbIdFileReadError(UsdbIdFileError):
    """Error reading file from file system"""

    def __str__(self) -> str:
        return "failed to read file"


@attrs.define
class UsdbIdFileInvalidFormatError(UsdbIdFileError):
    """Invalid file format"""

    def __str__(self) -> str:
        return "invalid file format"


@attrs.define
class UsdbIdFileMissingSectionHeaderFormatError(UsdbIdFileInvalidFormatError):
    """Invalid file format with missing section header"""

    def __str__(self) -> str:
        return f"{super().__str__()}: missing a section header"


@attrs.define
class UsdbIdFileMissingOrDuplicateOptionFormatError(UsdbIdFileInvalidFormatError):
    """Invalid file format with missing or duplicate option"""

    def __str__(self) -> str:
        return f"{super().__str__()}: missing or duplicate option"


@attrs.define
class UsdbIdFileMultipleUrlsFormatError(UsdbIdFileInvalidFormatError):
    """Invalid file format with multiple URLs"""

    def __str__(self) -> str:
        return f"{super().__str__()}: file contains multiple URLs"


@attrs.define
class UsdbIdFileMissingKeyFormatError(UsdbIdFileInvalidFormatError):
    """Invalid file format with missing key in file"""

    missing_key: str

    def __str__(self) -> str:
        return f"{super().__str__()}: missing key '{self.missing_key}'"


@attrs.define
class UsdbIdFileMissingSectionFormatError(UsdbIdFileInvalidFormatError):
    """Invalid file format with missing section in file"""

    missing_section: str

    def __str__(self) -> str:
        return f"{super().__str__()}: missing section '{self.missing_section}'"


@attrs.define
class UsdbIdFileMissingTagFormatError(UsdbIdFileInvalidFormatError):
    """Invalid file format with missing tag in file"""

    missing_tag: str

    def __str__(self) -> str:
        return f"{super().__str__()}: missing tag '{self.missing_tag}'"


@attrs.define
class UsdbIdFileMultipleTagsFormatError(UsdbIdFileInvalidFormatError):
    """Invalid file format with a tag occurring multiple times instead of just once"""

    multiple_tag: str

    def __str__(self) -> str:
        return f"{super().__str__()}: multiple tags '{self.multiple_tag}'"


@attrs.define
class UsdbIdFileMissingUrlTagFormatError(UsdbIdFileInvalidFormatError):
    """Invalid file format with missing tag for URL in file"""

    missing_tag: str

    def __str__(self) -> str:
        return f"{super().__str__()}: missing URL tag '{self.missing_tag}'"


@attrs.define
class UsdbIdFileMalformedUrlFormatError(UsdbIdFileInvalidFormatError):
    """Invalid file format with URL in bad format"""

    bad_url: str

    def __str__(self) -> str:
        return f"{super().__str__()}: malformed URL '{self.bad_url}'"


@attrs.define
class UsdbIdFileInvalidDomainMalformedUrlFormatError(UsdbIdFileMalformedUrlFormatError):
    """Invalid file format with URL containing bad domain"""

    bad_domain: str

    def __str__(self) -> str:
        return f"{super().__str__()}: has invalid domain '{self.bad_domain}'"


@attrs.define
class UsdbIdFileNoParametersMalformedUrlFormatError(UsdbIdFileMalformedUrlFormatError):
    """Invalid file format with URL having no query parameters"""

    def __str__(self) -> str:
        return f"{super().__str__()}: has no query parameters"


@attrs.define
class UsdbIdFileMissingQueryParameterMalformedUrlFormatError(
    UsdbIdFileMalformedUrlFormatError
):
    """Invalid file format with URL missing a specific query parameter"""

    missing_parameter: str

    def __str__(self) -> str:
        return (
            f"{super().__str__()}: missing query parameter '{self.missing_parameter}'"
        )


@attrs.define
class UsdbIdFileRepeatedQueryParameterMalformedUrlFormatError(
    UsdbIdFileMalformedUrlFormatError
):
    """Invalid file format with URL specific query parameter occurring multiple times"""

    repeated_parameter: str

    def __str__(self) -> str:
        return (
            f"{super().__str__()}: repeated query parameter '{self.repeated_parameter}'"
        )


@attrs.define
class UsdbIdFileInvalidQueryParameterMalformedUrlFormatError(
    UsdbIdFileMalformedUrlFormatError
):
    """Invalid file format with URL having an invalid query parameter"""

    invalid_parameter: str

    def __str__(self) -> str:
        return (
            f"{super().__str__()}: invalid query parameter '{self.invalid_parameter}'"
        )


@attrs.define
class UsdbIdFileUnparsableQueryParameterMalformedUrlFormatError(
    UsdbIdFileMalformedUrlFormatError
):
    """Invalid file format with URL having a query parameter that cannot be parsed"""

    unparsable_parameter: str

    def __str__(self) -> str:
        return (
            f"{super().__str__()}: could not parse query parameter "
            f"'{self.unparsable_parameter}'"
        )


@attrs.define
class UsdbIdFileEmptyFileError(UsdbIdFileError):
    """Files do not contain any USDB ID but were selected for import"""

    def __str__(self) -> str:
        return "empty file"


@attrs.define
class UsdbIdFileInvalidJsonError(UsdbIdFileError):
    """failed to interpret file content as JSON"""

    def __str__(self) -> str:
        return "invalid JSON format"


@attrs.define
class UsdbIdFileEmptySongsArrayError(UsdbIdFileError):
    """songs array is empty"""

    songs_key: str

    def __str__(self) -> str:
        return f"'{self.songs_key}' is empty"


@attrs.define
class UsdbIdFileWrongJsonSongsFormatError(UsdbIdFileError):
    """songs value is not an array"""

    songs_key: str

    def __str__(self) -> str:
        return f"'{self.songs_key}' is not a JSON array"


@attrs.define
class UsdbIdFileInvalidUsdbIdError(UsdbIdFileError):
    """expected USDB ID string cannot be converted correctly"""

    def __str__(self) -> str:
        return "invalid USDB ID in file"


@attrs.define
class UnexpectedUsdbIdFileInvalidUsdbIdError(UsdbIdFileInvalidUsdbIdError):
    """some unknown error around parsing an USDB ID string"""

    def __str__(self) -> str:
        return "unexpected error when parsing USDB ID(s)"


@attrs.define
class UsdbIdFileNoUrlFoundError(UsdbIdFileError):
    """parser could not find an URL in file content"""

    def __str__(self) -> str:
        return "no URL found"


def get_available_song_ids_from_files(file_list: list[Path]) -> list[SongId]:
    song_ids: list[SongId] = []
    for path in file_list:
        try:
            song_ids += parse_usdb_id_file(path)
        except UsdbIdFileError as error:
            logger.error(f"Failed to import file '{path}': {error!s}")
            return []

    unique_song_ids = list(set(song_ids))
    unique_song_ids.sort()
    logger.info(
        f"read {len(file_list)} file(s), "
        f"found {len(unique_song_ids)} "
        f"USDB IDs: {', '.join(str(id_) for id_ in unique_song_ids)}"
    )
    if unavailable_song_ids := [
        song_id for song_id in unique_song_ids if not UsdbSong.get(song_id)
    ]:
        logger.warning(
            f"{len(unavailable_song_ids)}/{len(unique_song_ids)} "
            "imported USDB IDs are not available: "
            f"{', '.join(str(song_id) for song_id in unavailable_song_ids)}"
        )

    if available_song_ids := [
        song_id for song_id in unique_song_ids if song_id not in unavailable_song_ids
    ]:
        logger.info(
            f"available {len(available_song_ids)}/{len(unique_song_ids)} "
            "imported USDB IDs will be selected: "
            f"{', '.join(str(song_id) for song_id in available_song_ids)}"
        )

    return available_song_ids


def _get_json_file_content(filepath: Path) -> str:
    filecontent: str
    try:
        with Path(filepath).open("r", encoding="utf-8") as file:
            filecontent = file.read()
    except OSError as exception:
        raise UsdbIdFileReadError() from exception
    except Exception as exception:
        raise UnexpectedUsdbIdFileError() from exception
    if not filecontent:
        raise UsdbIdFileEmptyFileError()
    return filecontent


def _parse_json_file(filepath: Path) -> list[SongId]:
    filecontent = _get_json_file_content(filepath)

    parsed_json = None
    try:
        parsed_json = json.loads(filecontent)
    except json.decoder.JSONDecodeError as exception:
        raise UsdbIdFileInvalidJsonError() from exception
    except Exception as exception:
        raise UnexpectedUsdbIdFileError() from exception

    if not isinstance(parsed_json, dict):
        raise UsdbIdFileInvalidJsonError()

    return _parse_json_content(parsed_json)


def _parse_json_content(parsed_json: dict) -> list[SongId]:
    top_key = "songs"

    if top_key not in parsed_json:
        raise UsdbIdFileMissingKeyFormatError(top_key)
    if not isinstance(parsed_json[top_key], list):
        raise UsdbIdFileWrongJsonSongsFormatError(songs_key=top_key)
    if not parsed_json[top_key]:
        raise UsdbIdFileEmptySongsArrayError(songs_key=top_key)

    key = "id"
    try:
        return [SongId.parse(element[key]) for element in parsed_json[top_key]]
    except ValueError as exception:
        raise UsdbIdFileInvalidUsdbIdError() from exception
    except (KeyError, IndexError) as exception:
        raise UsdbIdFileMissingKeyFormatError(exception.args[0]) from exception
    except Exception as exception:
        raise UnexpectedUsdbIdFileInvalidUsdbIdError() from exception


def _parse_ini_file(filepath: Path, section: str, key: str) -> SongId:
    config = configparser.ConfigParser()
    try:
        config.read(filepath)
    except configparser.MissingSectionHeaderError as exception:
        raise UsdbIdFileMissingSectionHeaderFormatError() from exception
    except Exception as exception:
        raise UsdbIdFileMissingOrDuplicateOptionFormatError() from exception
    if not config.sections():
        raise UsdbIdFileEmptyFileError()
    if section not in config:
        raise UsdbIdFileMissingSectionFormatError(section)
    if key not in config[section]:
        raise UsdbIdFileMissingKeyFormatError(key)
    url = config[section][key]
    return _parse_url(url)


def _parse_url_file(filepath: Path) -> SongId:
    return _parse_ini_file(filepath, section="InternetShortcut", key="URL")


def _parse_desktop_file(filepath: Path) -> SongId:
    return _parse_ini_file(filepath, section="Desktop Entry", key="URL")


def _get_soup(filepath: Path) -> BeautifulSoup:
    try:
        with Path(filepath).open("r", encoding="utf-8") as file:
            soup = BeautifulSoup(file, features="lxml-xml")
    except OSError as exception:
        raise UsdbIdFileReadError() from exception
    except Exception as exception:
        raise UnexpectedUsdbIdFileError() from exception

    if soup.is_empty_element:
        raise UsdbIdFileEmptyFileError()
    return soup


def _parse_webloc_file(filepath: Path) -> SongId:
    soup = _get_soup(filepath)
    tag = "plist"
    xml_plist = soup.find_all(tag)
    if not xml_plist or not isinstance(plist_tag := xml_plist[0], bs4.Tag):
        raise UsdbIdFileMissingTagFormatError(tag)
    if len(xml_plist) > 1:
        raise UsdbIdFileMultipleTagsFormatError(tag)
    tag = "dict"
    xml_dict = plist_tag.find_all(tag)
    if not xml_dict:
        raise UsdbIdFileMissingTagFormatError(tag)
    if len(xml_dict) > 1:
        raise UsdbIdFileMultipleTagsFormatError(tag)
    tag = "string"
    if not isinstance(dict_tag := xml_dict[0], bs4.Tag) or not (
        xml_string := dict_tag.find_all(tag)
    ):
        raise UsdbIdFileMissingUrlTagFormatError(tag)
    if len(xml_string) > 1:
        raise UsdbIdFileMultipleUrlsFormatError()

    url = xml_string[0].get_text()
    return _parse_url(url)


def _parse_usdb_ids_file(filepath: Path) -> list[SongId]:
    lines: list[str] = []
    try:
        with Path(filepath).open("r", encoding="utf-8") as file:
            lines = file.readlines()
    except OSError as exception:
        raise UsdbIdFileReadError() from exception
    except Exception as exception:
        raise UnexpectedUsdbIdFileError() from exception

    if not lines:
        raise UsdbIdFileEmptyFileError()

    try:
        return [SongId.parse(line) for line in lines]
    except ValueError as exception:
        raise UsdbIdFileInvalidUsdbIdError() from exception
    except Exception as exception:
        raise UnexpectedUsdbIdFileInvalidUsdbIdError() from exception


def _parse_url(url: str | None) -> SongId:
    if not url:
        raise UsdbIdFileNoUrlFoundError()
    parsed_url = urlparse(url)
    if not parsed_url.netloc:
        raise UsdbIdFileMalformedUrlFormatError(url)
    if parsed_url.netloc != "usdb.animux.de":
        raise UsdbIdFileInvalidDomainMalformedUrlFormatError(url, parsed_url.netloc)
    if not parsed_url.query:
        raise UsdbIdFileNoParametersMalformedUrlFormatError(url)
    query_params = parse_qs(parsed_url.query)
    id_param = "id"
    if id_param not in query_params:
        raise UsdbIdFileMissingQueryParameterMalformedUrlFormatError(url, id_param)
    if len(query_params[id_param]) > 1:
        raise UsdbIdFileRepeatedQueryParameterMalformedUrlFormatError(url, id_param)
    try:
        return SongId.parse(query_params[id_param][0])
    except ValueError as exception:
        raise UsdbIdFileInvalidQueryParameterMalformedUrlFormatError(
            url, id_param
        ) from exception
    except Exception as exception:
        # handle any other exception
        raise UsdbIdFileUnparsableQueryParameterMalformedUrlFormatError(
            url, id_param
        ) from exception


def parse_usdb_id_file(filepath: Path) -> list[SongId]:
    """parses files for USDB IDs"""
    song_ids: list[SongId] = []
    if filepath.suffix == ".json":
        song_ids = _parse_json_file(filepath)
    elif filepath.suffix == ".url":
        song_ids = [_parse_url_file(filepath)]
    elif filepath.suffix == ".desktop":
        song_ids = [_parse_desktop_file(filepath)]
    elif filepath.suffix == ".webloc":
        song_ids = [_parse_webloc_file(filepath)]
    elif filepath.suffix == ".usdb_ids":
        song_ids = _parse_usdb_ids_file(filepath)
    else:
        raise UsdbIdFileUnsupportedExtensionError()
    return song_ids


def write_usdb_id_file(filepath: Path, song_ids: Iterable[SongId]) -> None:
    with Path(filepath).open(encoding="utf-8", mode="w") as file:
        file.write("\n".join(str(id_) for id_ in song_ids))



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ReportDialog.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QComboBox, QDialog,
    QDialogButtonBox, QGridLayout, QGroupBox, QHBoxLayout,
    QLabel, QListWidget, QListWidgetItem, QRadioButton,
    QSizePolicy, QSpacerItem, QSpinBox, QSplitter,
    QTabWidget, QVBoxLayout, QWidget)
from usdb_syncer.gui.resources.qt import resources as resources_rc

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(663, 498)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Dialog.sizePolicy().hasHeightForWidth())
        Dialog.setSizePolicy(sizePolicy)
        Dialog.setMinimumSize(QSize(0, 0))
        self.verticalLayout_5 = QVBoxLayout(Dialog)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.groupBox_source = QGroupBox(Dialog)
        self.groupBox_source.setObjectName(u"groupBox_source")
        self.verticalLayout_3 = QVBoxLayout(self.groupBox_source)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.radioButton_locally_available_songs = QRadioButton(self.groupBox_source)
        self.radioButton_locally_available_songs.setObjectName(u"radioButton_locally_available_songs")
        self.radioButton_locally_available_songs.setChecked(True)

        self.verticalLayout_3.addWidget(self.radioButton_locally_available_songs)

        self.radioButton_selected_songs = QRadioButton(self.groupBox_source)
        self.radioButton_selected_songs.setObjectName(u"radioButton_selected_songs")

        self.verticalLayout_3.addWidget(self.radioButton_selected_songs)

        self.radioButton_all_songs = QRadioButton(self.groupBox_source)
        self.radioButton_all_songs.setObjectName(u"radioButton_all_songs")

        self.verticalLayout_3.addWidget(self.radioButton_all_songs)


        self.verticalLayout_5.addWidget(self.groupBox_source)

        self.tabWidget_report_type = QTabWidget(Dialog)
        self.tabWidget_report_type.setObjectName(u"tabWidget_report_type")
        self.tabWidget_report_type.setMinimumSize(QSize(0, 200))
        self.tab_pdf = QWidget()
        self.tab_pdf.setObjectName(u"tab_pdf")
        self.verticalLayout_2 = QVBoxLayout(self.tab_pdf)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.splitter = QSplitter(self.tab_pdf)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Orientation.Horizontal)
        self.layoutWidget = QWidget(self.splitter)
        self.layoutWidget.setObjectName(u"layoutWidget")
        self.verticalLayout_columns = QVBoxLayout(self.layoutWidget)
        self.verticalLayout_columns.setObjectName(u"verticalLayout_columns")
        self.verticalLayout_columns.setContentsMargins(0, 0, 0, 0)
        self.label_7 = QLabel(self.layoutWidget)
        self.label_7.setObjectName(u"label_7")

        self.verticalLayout_columns.addWidget(self.label_7)

        self.optional_columns = QListWidget(self.layoutWidget)
        self.optional_columns.setObjectName(u"optional_columns")

        self.verticalLayout_columns.addWidget(self.optional_columns)

        self.splitter.addWidget(self.layoutWidget)
        self.layoutWidget1 = QWidget(self.splitter)
        self.layoutWidget1.setObjectName(u"layoutWidget1")
        self.verticalLayout_settings = QVBoxLayout(self.layoutWidget1)
        self.verticalLayout_settings.setObjectName(u"verticalLayout_settings")
        self.verticalLayout_settings.setContentsMargins(0, 0, 0, 0)
        self.groupBox_page_layout = QGroupBox(self.layoutWidget1)
        self.groupBox_page_layout.setObjectName(u"groupBox_page_layout")
        self.gridLayout = QGridLayout(self.groupBox_page_layout)
        self.gridLayout.setObjectName(u"gridLayout")
        self.label_columns = QLabel(self.groupBox_page_layout)
        self.label_columns.setObjectName(u"label_columns")

        self.gridLayout.addWidget(self.label_columns, 3, 0, 1, 1)

        self.label_font_size = QLabel(self.groupBox_page_layout)
        self.label_font_size.setObjectName(u"label_font_size")

        self.gridLayout.addWidget(self.label_font_size, 4, 0, 1, 1)

        self.label_pagesize = QLabel(self.groupBox_page_layout)
        self.label_pagesize.setObjectName(u"label_pagesize")

        self.gridLayout.addWidget(self.label_pagesize, 0, 0, 1, 1)

        self.comboBox_pdf_orientation = QComboBox(self.groupBox_page_layout)
        self.comboBox_pdf_orientation.setObjectName(u"comboBox_pdf_orientation")

        self.gridLayout.addWidget(self.comboBox_pdf_orientation, 1, 1, 1, 1)

        self.spinBox_pdf_font_size = QSpinBox(self.groupBox_page_layout)
        self.spinBox_pdf_font_size.setObjectName(u"spinBox_pdf_font_size")
        self.spinBox_pdf_font_size.setMinimum(7)
        self.spinBox_pdf_font_size.setMaximum(20)
        self.spinBox_pdf_font_size.setValue(10)

        self.gridLayout.addWidget(self.spinBox_pdf_font_size, 4, 1, 1, 1)

        self.spinBox_pdf_columns = QSpinBox(self.groupBox_page_layout)
        self.spinBox_pdf_columns.setObjectName(u"spinBox_pdf_columns")
        self.spinBox_pdf_columns.setMinimum(1)
        self.spinBox_pdf_columns.setMaximum(4)

        self.gridLayout.addWidget(self.spinBox_pdf_columns, 3, 1, 1, 1)

        self.label_orientation = QLabel(self.groupBox_page_layout)
        self.label_orientation.setObjectName(u"label_orientation")

        self.gridLayout.addWidget(self.label_orientation, 1, 0, 1, 1)

        self.comboBox_pdf_pagesize = QComboBox(self.groupBox_page_layout)
        self.comboBox_pdf_pagesize.setObjectName(u"comboBox_pdf_pagesize")

        self.gridLayout.addWidget(self.comboBox_pdf_pagesize, 0, 1, 1, 1)

        self.spinBox_pdf_margin = QSpinBox(self.groupBox_page_layout)
        self.spinBox_pdf_margin.setObjectName(u"spinBox_pdf_margin")
        self.spinBox_pdf_margin.setMinimum(10)
        self.spinBox_pdf_margin.setMaximum(30)
        self.spinBox_pdf_margin.setSingleStep(5)
        self.spinBox_pdf_margin.setValue(20)

        self.gridLayout.addWidget(self.spinBox_pdf_margin, 2, 1, 1, 1)

        self.label_margin = QLabel(self.groupBox_page_layout)
        self.label_margin.setObjectName(u"label_margin")

        self.gridLayout.addWidget(self.label_margin, 2, 0, 1, 1)

        self.verticalSpacer_3 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer_3, 5, 1, 1, 1)


        self.verticalLayout_settings.addWidget(self.groupBox_page_layout)

        self.splitter.addWidget(self.layoutWidget1)

        self.verticalLayout_2.addWidget(self.splitter)

        self.tabWidget_report_type.addTab(self.tab_pdf, "")
        self.tab_json = QWidget()
        self.tab_json.setObjectName(u"tab_json")
        self.verticalLayout_4 = QVBoxLayout(self.tab_json)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label_json_indent = QLabel(self.tab_json)
        self.label_json_indent.setObjectName(u"label_json_indent")

        self.horizontalLayout.addWidget(self.label_json_indent)

        self.spinBox_json_indent = QSpinBox(self.tab_json)
        self.spinBox_json_indent.setObjectName(u"spinBox_json_indent")
        self.spinBox_json_indent.setMinimum(2)
        self.spinBox_json_indent.setMaximum(4)
        self.spinBox_json_indent.setSingleStep(2)
        self.spinBox_json_indent.setValue(4)

        self.horizontalLayout.addWidget(self.spinBox_json_indent)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)


        self.verticalLayout_4.addLayout(self.horizontalLayout)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_4.addItem(self.verticalSpacer_2)

        self.tabWidget_report_type.addTab(self.tab_json, "")

        self.verticalLayout_5.addWidget(self.tabWidget_report_type)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_5.addItem(self.verticalSpacer)

        self.buttonBox = QDialogButtonBox(Dialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)

        self.verticalLayout_5.addWidget(self.buttonBox)


        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)

        self.tabWidget_report_type.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"New Report", None))
        self.groupBox_source.setTitle(QCoreApplication.translate("Dialog", u"Source", None))
        self.radioButton_locally_available_songs.setText(QCoreApplication.translate("Dialog", u"Locally available songs", None))
        self.radioButton_selected_songs.setText(QCoreApplication.translate("Dialog", u"Selected songs", None))
        self.radioButton_all_songs.setText(QCoreApplication.translate("Dialog", u"All songs", None))
        self.label_7.setText(QCoreApplication.translate("Dialog", u"Optional data to include:", None))
        self.groupBox_page_layout.setTitle(QCoreApplication.translate("Dialog", u"PDF Page layout", None))
        self.label_columns.setText(QCoreApplication.translate("Dialog", u"Columns:", None))
        self.label_font_size.setText(QCoreApplication.translate("Dialog", u"Font size:", None))
        self.label_pagesize.setText(QCoreApplication.translate("Dialog", u"Page size:", None))
        self.label_orientation.setText(QCoreApplication.translate("Dialog", u"Orientation:", None))
        self.label_margin.setText(QCoreApplication.translate("Dialog", u"Margins:", None))
        self.tabWidget_report_type.setTabText(self.tabWidget_report_type.indexOf(self.tab_pdf), QCoreApplication.translate("Dialog", u"PDF Report", None))
        self.label_json_indent.setText(QCoreApplication.translate("Dialog", u"Indent:", None))
        self.tabWidget_report_type.setTabText(self.tabWidget_report_type.indexOf(self.tab_json), QCoreApplication.translate("Dialog", u"JSON Report", None))
    # retranslateUi




"""Hooks that add-ons can subscribe to."""

from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Generic, ParamSpec

import attrs

from usdb_syncer.logger import logger

if TYPE_CHECKING:
    from usdb_syncer import usdb_song


P = ParamSpec("P")


@attrs.define(slots=False)
class _Hook(Generic[P]):
    """Base class for hooks."""

    _subscribers: list[Callable[P, None]]

    def __init_subclass__(cls) -> None:
        cls._subscribers = []

    @classmethod
    def subscribe(cls, func: Callable[P, None]) -> None:
        cls._subscribers.append(func)

    @classmethod
    def unsubscribe(cls, func: Callable[P, None]) -> None:
        cls._subscribers.remove(func)

    @classmethod
    def call(cls, *args: P.args, **kwargs: P.kwargs) -> None:
        for func in cls._subscribers:
            try:
                func(*args, **kwargs)
            except Exception as e:  # noqa: BLE001
                logger.exception(
                    f"Plugin error in {func.__name__}: {type(e).__name__}: {e}"
                )


class SongLoaderDidFinish(_Hook):
    """Called after downloading a song."""

    @classmethod
    def call(cls, song: usdb_song.UsdbSong) -> None:
        super().call(song)



"""Representations of rows of the filter tree."""

from __future__ import annotations

import enum
from collections.abc import Iterable
from typing import Any, Generic, TypeVar, assert_never

import attrs
from PySide6.QtCore import Qt
from PySide6.QtGui import QIcon

from usdb_syncer import db
from usdb_syncer.gui.icons import Icon


class SongMatch:
    """Interface for objects that can be matched against a song."""

    def build_search(self, search: db.SearchBuilder) -> None:
        raise NotImplementedError

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        raise NotImplementedError

    def is_accepted(self, _matches: set[str | int]) -> bool:
        return True


T = TypeVar("T")


@attrs.define
class SongValueMatch(SongMatch, Generic[T]):
    """str that can be matched against a specific attribute of a song."""

    val: T
    count: int

    def __str__(self) -> str:
        return f"{self.val} [{self.count}]"

    def search_attr(self, search: db.SearchBuilder) -> list[T]:
        raise NotImplementedError

    def build_search(self, search: db.SearchBuilder) -> None:
        self.search_attr(search).append(self.val)

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        return self.val in self.search_attr(search)

    def is_accepted(self, matches: set[str | int]) -> bool:
        return self.val in matches


class SongArtistMatch(SongValueMatch):
    """str that can be matched against a song's artist."""

    def search_attr(self, search: db.SearchBuilder) -> list[str]:
        return search.artists


class SongTitleMatch(SongValueMatch):
    """str that can be matched against a song's title."""

    def search_attr(self, search: db.SearchBuilder) -> list[str]:
        return search.titles


class SongEditionMatch(SongValueMatch):
    """str that can be matched against a song's edition."""

    def search_attr(self, search: db.SearchBuilder) -> list[str]:
        return search.editions


class SongLanguageMatch(SongValueMatch):
    """str that can be matched against a song's language."""

    def search_attr(self, search: db.SearchBuilder) -> list[str]:
        return search.languages


class SongYearMatch(SongValueMatch):
    """str that can be matched against a song's year."""

    def search_attr(self, search: db.SearchBuilder) -> list[int]:
        return search.years


class SongGenreMatch(SongValueMatch):
    """str that can be matched against a song's genre."""

    def search_attr(self, search: db.SearchBuilder) -> list[str]:
        return search.genres


class SongCreatorMatch(SongValueMatch):
    """str that can be matched against a song's creator."""

    def search_attr(self, search: db.SearchBuilder) -> list[str]:
        return search.creators


@attrs.define(kw_only=True)
class TreeItem:
    """A row in the tree."""

    data: Any
    parent: TreeItem | None
    checkable: bool = False
    row_in_parent: int = attrs.field(default=0, init=False)
    children: tuple[TreeItem, ...] = attrs.field(factory=tuple, init=False)
    checked: bool | None = attrs.field(default=False, init=False)

    def toggle_checked(self, _keep_siblings: bool) -> tuple[TreeItem, ...]:
        """Returns toggled items."""
        return ()

    def decoration(self) -> QIcon | None:
        return None

    def flags(self) -> Qt.ItemFlag:
        return Qt.ItemFlag.ItemIsEnabled

    def is_accepted(self, _matches: dict[Filter, set[str | int]]) -> bool:
        return True


@attrs.define(kw_only=True)
class RootItem(TreeItem):
    """The root item of the tree. Only used internally."""

    data: None = attrs.field(default=None, init=False)
    parent: None = attrs.field(default=None, init=False)
    children: tuple[FilterItem, ...] = attrs.field(factory=tuple, init=False)

    def set_children(self, children: Iterable[FilterItem]) -> None:
        self.children = tuple(children)
        for row, child in enumerate(self.children):
            child.parent = self
            child.row_in_parent = row


@attrs.define(kw_only=True)
class FilterItem(TreeItem):
    """A top-level row in the tree representing a filter kind."""

    data: Filter
    parent: RootItem
    children: tuple[VariantItem, ...] = attrs.field(factory=tuple, init=False)
    checked_children: set[int] = attrs.field(factory=set, init=False)

    def __attrs_post_init__(self) -> None:
        self.checkable = self.data != Filter.SAVED

    def add_child(self, child: VariantItem) -> None:
        child.parent = self
        child.row_in_parent = len(self.children)
        self.children = (*self.children, child)

    def set_children(self, children: Iterable[VariantItem]) -> None:
        self.children = tuple(children)
        for row, child in enumerate(self.children):
            child.parent = self
            child.row_in_parent = row

    def flags(self) -> Qt.ItemFlag:
        if self.checked_children:
            return Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsUserCheckable
        return Qt.ItemFlag.ItemIsEnabled

    def toggle_checked(self, _keep_siblings: bool) -> tuple[TreeItem, ...]:
        return self.uncheck_children() if self.checked else ()

    def uncheck_children(self) -> tuple[TreeItem, ...]:
        changed = tuple(self.checked_child_items())
        for child in changed:
            child.checked = False
        self.checked_children.clear()
        self.checked = False
        return (*changed, self)

    def set_child_checked(
        self, child: int, checked: bool, keep_siblings: bool
    ) -> tuple[TreeItem, ...]:
        changed: tuple[TreeItem, ...] = (self, self.children[child])
        if checked:
            if not keep_siblings:
                changed += self.uncheck_children()
            self.checked_children.add(child)
        else:
            self.checked_children.remove(child)
        self.children[child].checked = checked
        self.checked = bool(self.checked_children)
        return changed

    def set_checked_children(self, search: db.SearchBuilder) -> list[TreeItem]:
        changed: list[TreeItem] = []
        for idx, child in enumerate(self.children):
            if child.data.is_in_search(search) == child.checked:
                continue
            changed.append(child)
            child.checked = not child.checked
            if child.checked:
                self.checked_children.add(idx)
            else:
                self.checked_children.remove(idx)
        if self.checked != bool(self.checked_children):
            changed.append(self)
            self.checked = not self.checked
        return changed

    def decoration(self) -> QIcon:
        return self.data.decoration()

    def checked_child_items(self) -> Iterable[VariantItem]:
        return (self.children[row] for row in self.checked_children)

    def build_search(self, search: db.SearchBuilder) -> None:
        for child in self.checked_child_items():
            child.data.build_search(search)


@attrs.define(kw_only=True)
class VariantItem(TreeItem):
    """A node row in the tree representing one variant of a specific filter."""

    data: SongMatch
    parent: FilterItem
    children: tuple[TreeItem, ...] = attrs.field(factory=tuple, init=False)
    _flags: Qt.ItemFlag = attrs.field(init=False)

    def __attrs_post_init__(self) -> None:
        if self.parent.data == Filter.SAVED:
            self.checkable = False
            self._flags = (
                Qt.ItemFlag.ItemIsEnabled
                | Qt.ItemFlag.ItemNeverHasChildren
                | Qt.ItemFlag.ItemIsEditable
            )
        else:
            self.checkable = True
            self._flags = (
                Qt.ItemFlag.ItemIsEnabled
                | Qt.ItemFlag.ItemIsUserCheckable
                | Qt.ItemFlag.ItemNeverHasChildren
            )

    def flags(self) -> Qt.ItemFlag:
        return self._flags

    def toggle_checked(self, keep_siblings: bool) -> tuple[TreeItem, ...]:
        return self.parent.set_child_checked(
            self.row_in_parent, not self.checked, keep_siblings
        )

    def is_accepted(self, matches: dict[Filter, set[str | int]]) -> bool:
        if (parent_matches := matches.get(self.parent.data)) is None:
            return True
        return self.data.is_accepted(parent_matches)


class Filter(enum.Enum):
    """Kinds of filters in the tree."""

    SAVED = 0
    STATUS = enum.auto()
    ARTIST = enum.auto()
    TITLE = enum.auto()
    EDITION = enum.auto()
    LANGUAGE = enum.auto()
    GOLDEN_NOTES = enum.auto()
    RATING = enum.auto()
    VIEWS = enum.auto()
    YEAR = enum.auto()
    GENRE = enum.auto()
    CREATOR = enum.auto()

    def __str__(self) -> str:  # noqa: C901
        match self:
            case Filter.SAVED:
                return "Saved Searches"
            case Filter.STATUS:
                return "Status"
            case Filter.ARTIST:
                return "Artist"
            case Filter.TITLE:
                return "Title"
            case Filter.EDITION:
                return "Edition"
            case Filter.LANGUAGE:
                return "Language"
            case Filter.GOLDEN_NOTES:
                return "Golden Notes"
            case Filter.RATING:
                return "Rating"
            case Filter.VIEWS:
                return "Views"
            case Filter.YEAR:
                return "Year"
            case Filter.GENRE:
                return "Genre"
            case Filter.CREATOR:
                return "Creator"
            case _ as unreachable:
                assert_never(unreachable)

    def variants(self) -> Iterable[SongMatch]:  # noqa: C901
        match self:
            case Filter.SAVED:
                return SavedSearch.load_all()
            case Filter.ARTIST:
                return (SongArtistMatch(v, c) for v, c in db.usdb_song_artists())
            case Filter.TITLE:
                return (SongTitleMatch(v, c) for v, c in db.usdb_song_titles())
            case Filter.EDITION:
                return (SongEditionMatch(v, c) for v, c in db.usdb_song_editions())
            case Filter.LANGUAGE:
                return (SongLanguageMatch(v, c) for v, c in db.usdb_song_languages())
            case Filter.STATUS:
                return StatusVariant
            case Filter.GOLDEN_NOTES:
                return GoldenNotesVariant
            case Filter.RATING:
                return RatingVariant
            case Filter.VIEWS:
                return ViewsVariant
            case Filter.YEAR:
                return (SongYearMatch(v, c) for v, c in db.usdb_song_years())
            case Filter.GENRE:
                return (SongGenreMatch(v, c) for v, c in db.usdb_song_genres())
            case Filter.CREATOR:
                return (SongCreatorMatch(v, c) for v, c in db.usdb_song_creators())
            case _ as unreachable:
                assert_never(unreachable)

    def decoration(self) -> QIcon:  # noqa: C901
        match self:
            case Filter.SAVED:
                icon = Icon.SAVED_SEARCH
            case Filter.STATUS:
                icon = Icon.DOWNLOAD
            case Filter.ARTIST:
                icon = Icon.ARTIST
            case Filter.TITLE:
                icon = Icon.TITLE
            case Filter.EDITION:
                icon = Icon.EDITION
            case Filter.LANGUAGE:
                icon = Icon.LANGUAGE
            case Filter.GOLDEN_NOTES:
                icon = Icon.GOLDEN_NOTES
            case Filter.RATING:
                icon = Icon.RATING
            case Filter.VIEWS:
                icon = Icon.VIEWS
            case Filter.YEAR:
                icon = Icon.CALENDAR
            case Filter.GENRE:
                icon = Icon.GENRE
            case Filter.CREATOR:
                icon = Icon.CREATOR
            case _ as unreachable:
                assert_never(unreachable)
        return icon.icon()


class StatusVariant(SongMatch, enum.Enum):
    """Variants of the status of a song."""

    NONE = enum.auto()
    DOWNLOADED = enum.auto()
    IN_PROGRESS = enum.auto()
    FAILED = enum.auto()

    def __str__(self) -> str:
        match self:
            case StatusVariant.NONE:
                return "Not downloaded"
            case StatusVariant.DOWNLOADED:
                return "Downloaded"
            case StatusVariant.IN_PROGRESS:
                return "In progress"
            case StatusVariant.FAILED:
                return "Failed"
            case _ as unreachable:
                assert_never(unreachable)

    def build_search(self, search: db.SearchBuilder) -> None:
        match self:
            case StatusVariant.IN_PROGRESS:
                search.statuses.append(db.DownloadStatus.PENDING)
                search.statuses.append(db.DownloadStatus.DOWNLOADING)
            case StatusVariant.FAILED:
                search.statuses.append(db.DownloadStatus.FAILED)
            case StatusVariant.NONE | StatusVariant.DOWNLOADED:
                search.downloaded = (
                    self is StatusVariant.DOWNLOADED
                    if search.downloaded is None
                    else None
                )
            case unreachable:
                assert_never(unreachable)

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        match self:
            case StatusVariant.IN_PROGRESS:
                return (
                    db.DownloadStatus.PENDING in search.statuses
                    and db.DownloadStatus.DOWNLOADING in search.statuses
                )
            case StatusVariant.FAILED:
                return db.DownloadStatus.FAILED in search.statuses
            case StatusVariant.NONE:
                return search.downloaded is False
            case StatusVariant.DOWNLOADED:
                return search.downloaded is True
            case unreachable:
                assert_never(unreachable)


class RatingVariant(SongMatch, enum.Enum):
    """Selectable variants for the song rating filter."""

    R_NONE = None
    R_1 = 1
    R_2 = 2
    R_3 = 3
    R_4 = 4
    R_5 = 5

    def __str__(self) -> str:
        if self == RatingVariant.R_NONE:
            return "None"
        return self.value * "★"

    def build_search(self, search: db.SearchBuilder) -> None:
        search.ratings.append(self.value or 0)

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        return (self.value or 0) in search.ratings


class GoldenNotesVariant(SongMatch, enum.Enum):
    """Selectable variants for the golden notes filter."""

    NO = False
    YES = True

    def __str__(self) -> str:
        match self:
            case GoldenNotesVariant.NO:
                return "No"
            case GoldenNotesVariant.YES:
                return "Yes"
            case _ as unreachable:
                assert_never(unreachable)

    def build_search(self, search: db.SearchBuilder) -> None:
        if search.golden_notes is None:
            search.golden_notes = self.value
        else:
            # Yes *and* No is the same as no filter
            search.golden_notes = None

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        return self.value == search.golden_notes


class ViewsVariant(SongMatch, enum.Enum):
    """Selectable variants for the views filter."""

    V_0 = (0, 100)
    V_100 = (100, 200)
    V_200 = (200, 300)
    V_300 = (300, 400)
    V_400 = (400, 500)
    V_500 = (500, None)

    def __str__(self) -> str:
        if self.value[1] is None:
            return f"{self.value[0]}+"
        return f"{self.value[0]} to {self.value[1] - 1}"

    def build_search(self, search: db.SearchBuilder) -> None:
        search.views.append(self.value)

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        return self.value in search.views


@attrs.define
class SavedSearch(SongMatch, db.SavedSearch):
    """A search saved by the user."""

    @classmethod
    def load_all(cls) -> Iterable[SavedSearch]:
        with db.transaction():
            searches = db.SavedSearch.load_saved_searches()
        return (cls(s.name, s.search, s.is_default, s.subscribed) for s in searches)

    def build_search(self, search: db.SearchBuilder) -> None:
        pass

    def __str__(self) -> str:
        return self.name

    def is_in_search(self, search: db.SearchBuilder) -> bool:
        return False



"""Dialog to report missing ffmpeg."""

import sys
from collections.abc import Callable
from pathlib import Path

from PySide6.QtWidgets import QDialog, QFileDialog, QWidget

from usdb_syncer import settings
from usdb_syncer.gui.forms.FfmpegDialog import Ui_Dialog
from usdb_syncer.utils import add_to_system_path


def check_ffmpeg(parent: QWidget, on_success: Callable[[], None]) -> None:
    """If ffmpeg is available, can be restored from the settings or is provided
    by the user, executes `on_sucess`.
    """
    if settings.ffmpeg_is_available():
        on_success()
    else:
        FfmpegDialog(parent, on_success).show()


class FfmpegDialog(Ui_Dialog, QDialog):
    """Dialog to report missing ffmpeg."""

    def __init__(self, parent: QWidget, on_success: Callable[[], None]) -> None:
        self.on_success = on_success
        super().__init__(parent=parent)
        self.setupUi(self)
        match sys.platform:
            case "win32":
                self.label_windows.setVisible(True)
                self.label_macos.setVisible(False)
                self.label_linux.setVisible(False)
                self.setFixedHeight(194)
            case "darwin":
                self.label_windows.setVisible(False)
                self.label_macos.setVisible(True)
                self.label_linux.setVisible(False)
                self.setFixedHeight(140)
            case "linux" | "linux2":
                self.label_windows.setVisible(False)
                self.label_macos.setVisible(False)
                self.label_linux.setVisible(True)
                self.setFixedHeight(140)
            case _:
                pass
        self.set_location.clicked.connect(self._set_location)

    def _set_location(self) -> None:
        if not (path := self._get_ffmpeg_dir()):
            return
        add_to_system_path(path)
        settings.set_ffmpeg_dir(path)
        self.accept()
        self.on_success()

    def _get_ffmpeg_dir(self) -> str:
        filt = "ffmpeg.exe" if sys.platform == "win32" else "ffmpeg"
        path = QFileDialog.getOpenFileName(self, "Select ffmpeg", "", filt)[0]
        return str(Path(path).parent)



"""Dialog to post a comment on one or multiple songs."""

from PySide6.QtWidgets import QDialog, QWidget

from usdb_syncer.gui.forms.CommentDialog import Ui_Dialog
from usdb_syncer.usdb_scraper import post_song_comment
from usdb_syncer.usdb_song import UsdbSong


class CommentDialog(Ui_Dialog, QDialog):
    """Dialog to post a comment on one or multiple songs."""

    def __init__(self, parent: QWidget, selected_song: UsdbSong) -> None:
        super().__init__(parent=parent)
        self._parent = parent
        self.setupUi(self)
        self._selected_song = selected_song
        self.combobox_rating.addItem("Neutral", "neutral")
        self.combobox_rating.addItem("Positive", "positiv")
        self.combobox_rating.addItem("Negative", "negativ")

    def accept(self) -> None:
        song_id = self._selected_song.song_id
        text = self.text_edit_comment.toPlainText()
        rating = self.combobox_rating.currentData()
        post_song_comment(song_id, text, rating)
        super().accept()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'DebugConsole.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDialog, QPlainTextEdit, QSizePolicy,
    QSplitter, QVBoxLayout, QWidget)

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(659, 664)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.splitter = QSplitter(Dialog)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Vertical)
        self.splitter.setHandleWidth(5)
        self.input = QPlainTextEdit(self.splitter)
        self.input.setObjectName(u"input")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.input.sizePolicy().hasHeightForWidth())
        self.input.setSizePolicy(sizePolicy)
        font = QFont()
        font.setPointSize(10)
        self.input.setFont(font)
        self.input.setLineWidth(1)
        self.input.setPlaceholderText(u"This REPL lets you execute Python code at runtime.\n"
"\n"
"Actions\n"
"    Ctrl+Return         Execute\n"
"    Ctrl+Shift+Return   Execute and print current line\n"
"    Ctrl+L              Clear output\n"
"    Ctrl+Shift+L        Clear input\n"
"\n"
"Locals\n"
"    mw: usdb_syncer.gui.mw.MainWindow")
        self.splitter.addWidget(self.input)
        self.output = QPlainTextEdit(self.splitter)
        self.output.setObjectName(u"output")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(2)
        sizePolicy1.setHeightForWidth(self.output.sizePolicy().hasHeightForWidth())
        self.output.setSizePolicy(sizePolicy1)
        self.output.setFont(font)
        self.output.setReadOnly(True)
        self.splitter.addWidget(self.output)

        self.verticalLayout.addWidget(self.splitter)


        self.retranslateUi(Dialog)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Debug Console", None))
        self.input.setDocumentTitle("")
        self.input.setPlainText("")
    # retranslateUi




"""Errors thrown by the USDB Syncer."""


class UsdbSyncerError(Exception):
    """Common base class for our own errors."""


# common


class SongIdError(UsdbSyncerError):
    """Raised when a song id is invalid."""

    def __init__(self, value: int) -> None:
        super().__init__(f"Song id out of range: {value}")
        self.value = value


# database


class DatabaseError(UsdbSyncerError):
    """Base class for database errors."""


class AlreadyConnectedError(DatabaseError):
    """Raised when trying to connect to a database that is already connected."""


class NotConnectedError(DatabaseError):
    """Raised when trying to access a database that is not connected."""


class UnknownSchemaError(DatabaseError):
    """Raised if schema version is not compatible."""


# meta files


class MetaFileTooNewError(UsdbSyncerError):
    """Raised when trying to decode meta info from an incompatible future release."""

    def __str__(self) -> str:
        return "Cannot read sync meta written by a future release!"


# USDB


class UsdbError(UsdbSyncerError):
    """Super class for errors relating to USDB."""


class UsdbParseError(UsdbError):
    """Raised when HTML from USDB has unexpected format."""


class UsdbLoginError(UsdbError):
    """Raised when login was required, but not possible."""


class UsdbNotFoundError(UsdbError):
    """Raised when a requested USDB record is missing."""


class UsdbUnknownLanguageError(UsdbError):
    """Raised when the language of the USDB website cannot be determined."""


# txt parsing


class TxtParseError(UsdbSyncerError):
    """Raised when parsing a txt file fails."""


class HeadersParseError(TxtParseError):
    """Raised when failing to parse headers."""


class HeadersRequiredMissingError(HeadersParseError):
    """Raised when specific headers are required, but missing."""


class TrackParseError(TxtParseError):
    """Raised when failing to parse track."""


class InvalidCharError(TrackParseError):
    """Raised when a track contains invalid characters."""

    def __init__(self, type_: str, value: str) -> None:
        super().__init__(f"Invalid {type_}: {value}")
        self.type_ = type_
        self.value = value


class InvalidNoteError(InvalidCharError):
    """Raised when a note is invalid."""

    def __init__(self, note: str) -> None:
        super().__init__("note", note)
        self.note = note


class InvalidLineBreakError(InvalidCharError):
    """Raised when a line break is invalid."""

    def __init__(self, line_break: str) -> None:
        super().__init__("line break", line_break)
        self.line_break = line_break


class InvalidTrackError(TrackParseError):
    """Raised when a track is invalid."""

    def __init__(self) -> None:
        super().__init__("Invalid track.")


# user input


class AbortError(UsdbSyncerError):
    """Raised when the user requests to abort an operation."""



"""Signals other components can notify and subscribe to."""

from collections.abc import Callable
from pathlib import Path
from typing import Any, Self, cast

import attrs
from PySide6 import QtCore

from usdb_syncer import SongId, db, settings


class _EventProcessor(QtCore.QObject):
    """Processes events."""

    def customEvent(self, event: QtCore.QEvent) -> None:  # noqa: N802
        if isinstance(event, SubscriptableEvent):
            event.process()


class _EventProcessorManager(QtCore.QObject):
    """Singleton to manage an _EventProcessor instance."""

    _processor: _EventProcessor | None = None

    @classmethod
    def processor(cls) -> _EventProcessor:
        if cls._processor is None:
            cls._processor = _EventProcessor()
            app = QtCore.QCoreApplication.instance()
            assert app
            cls._processor.moveToThread(app.thread())
        return cls._processor


@attrs.define(slots=False)
class SubscriptableEvent(QtCore.QEvent):
    """An event other components can send and subscribe to."""

    _subscribers: list[Callable[[Self], Any]] = attrs.field(init=False)

    def __attrs_pre_init__(self) -> None:
        super().__init__(QtCore.QEvent.Type.User)

    def __init_subclass__(cls) -> None:
        cls._subscribers = []

    @classmethod
    def subscribe(cls, callback: Callable[[Self], Any]) -> None:
        # mypy seems to have a lot of trouble with the Self type
        cast(list[Callable[[Self], Any]], cls._subscribers).append(callback)

    @classmethod
    def unsubscribe(cls, callback: Callable[[Self], Any]) -> None:
        cast(list[Callable[[Self], Any]], cls._subscribers).remove(callback)

    def post(self) -> None:
        QtCore.QCoreApplication.postEvent(_EventProcessorManager.processor(), self)

    def process(self: Self) -> None:
        for func in self._subscribers:
            func(self)


# search


@attrs.define(slots=False)
class TreeFilterChanged(SubscriptableEvent):
    """Sent when a tree filter row has been selected or deselected."""

    search: db.SearchBuilder


@attrs.define(slots=False)
class TextFilterChanged(SubscriptableEvent):
    """Sent when the free text search has been changed."""

    search: str


@attrs.define(slots=False)
class SearchOrderChanged(SubscriptableEvent):
    """Sent when the search order has been changed or reversed."""

    order: db.SongOrder
    descending: bool


@attrs.define(slots=False)
class SavedSearchRestored(SubscriptableEvent):
    """Sent when the a save search is set."""

    search: db.SearchBuilder


# songs


@attrs.define(slots=False)
class SongChanged(SubscriptableEvent):
    """Sent when attributes of a UsdbSong have changed."""

    song_id: SongId


@attrs.define(slots=False)
class SongDeleted(SubscriptableEvent):
    """Sent when attributes of a UsdbSong have changed."""

    song_id: SongId


# downloads


@attrs.define(slots=False)
class DownloadsRequested(SubscriptableEvent):
    """Sent when a song download has been queued for download."""

    count: int


@attrs.define(slots=False)
class DownloadFinished(SubscriptableEvent):
    """Sent when a song download has finished."""

    song_id: SongId


# files


@attrs.define(slots=False)
class SongDirChanged(SubscriptableEvent):
    """Sent when the selected song directory has changed."""

    new_dir: Path


# UI


@attrs.define(slots=False)
class ThemeChanged(SubscriptableEvent):
    """Sent when a new theme has been applied."""

    theme: settings.Theme



"""Dialog to manage USDB login."""

from PySide6.QtGui import QDesktopServices
from PySide6.QtWidgets import QDialog, QMessageBox, QWidget

from usdb_syncer import settings
from usdb_syncer.constants import Usdb
from usdb_syncer.gui import icons
from usdb_syncer.gui.forms.UsdbLoginDialog import Ui_Dialog
from usdb_syncer.usdb_scraper import (
    SessionManager,
    get_logged_in_usdb_user,
    log_out_of_usdb,
    login_to_usdb,
    new_session_with_cookies,
)


class UsdbLoginDialog(Ui_Dialog, QDialog):
    """Dialog to manage USDB login."""

    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent=parent)
        self._parent = parent
        self.setupUi(self)
        self.command_link_register.pressed.connect(
            lambda: QDesktopServices.openUrl(Usdb.REGISTER_URL)
        )
        self.button_check_login.pressed.connect(self._on_check_login)
        self.button_log_out.pressed.connect(self._on_log_out)
        self._load_settings()

    def _load_settings(self) -> None:
        for browser in settings.Browser:
            if icon := icons.browser_icon(browser):
                self.combobox_browser.addItem(icon, str(browser), browser)
            else:
                self.combobox_browser.addItem(str(browser), browser)
        self.combobox_browser.setCurrentIndex(
            self.combobox_browser.findData(settings.get_browser())
        )
        user, password = settings.get_usdb_auth()
        self.line_edit_username.setText(user)
        self.line_edit_password.setText(password)

    def accept(self) -> None:
        settings.set_browser(self.combobox_browser.currentData())
        settings.set_usdb_auth(
            self.line_edit_username.text(), self.line_edit_password.text()
        )
        SessionManager.reset_session()
        super().accept()

    def _on_check_login(self) -> None:
        session = new_session_with_cookies(self.combobox_browser.currentData())
        if user := get_logged_in_usdb_user(session):
            message = f"Success! Existing session found with user '{user}'."
        elif (user := self.line_edit_username.text()) and (
            password := self.line_edit_password.text()
        ):
            if login_to_usdb(session, user, password):
                message = "Success! Logged in to USDB."
            else:
                message = "Login failed!"
        else:
            message = "No existing session found!"
        QMessageBox.information(self._parent, "Login Result", message)

    def _on_log_out(self) -> None:
        log_out_of_usdb(new_session_with_cookies(self.combobox_browser.currentData()))



"""General-purpose utilities."""

import datetime
import functools
import itertools
import os
import re
import subprocess
import sys
import time
import unicodedata
from pathlib import Path
from types import TracebackType
from typing import ClassVar

import requests
from appdirs import AppDirs
from bs4 import BeautifulSoup, Tag
from packaging import version
from unidecode import unidecode

import usdb_syncer
from usdb_syncer import constants
from usdb_syncer.logger import logger

CACHE_LIFETIME = 60 * 60
_app_dirs = AppDirs("usdb_syncer", "bohning")


# https://pyinstaller.org/en/stable/runtime-information.html#run-time-information
IS_BUNDLE = bool(getattr(sys, "frozen", False) and getattr(sys, "_MEIPASS", False))
IS_INSTALLED = "site-packages" in __file__
IS_SOURCE = not IS_BUNDLE and not IS_INSTALLED


def _root() -> Path:
    """Returns source root folder or temprory bundle folder if running as such."""
    if getattr(sys, "frozen", False) and (bundle := getattr(sys, "_MEIPASS", None)):
        return Path(bundle)
    return Path(__file__).parent.parent.parent.absolute()


def video_url_from_resource(resource: str) -> str | None:
    if "://" in resource:
        return resource
    if "/" in resource:
        return f"https://{resource}"
    vimeo_id_pattern = r"^\d{2,10}$"
    if re.match(vimeo_id_pattern, resource):
        return f"https://vimeo.com/{resource}"
    yt_id_pattern = r"^[A-Za-z0-9_-]{11}$"
    if re.match(yt_id_pattern, resource):
        return f"https://www.youtube.com/watch?v={resource}"
    return None


def get_allowed_countries(resource: str) -> list[str] | None:
    """Fetches YouTube video availability information from polsy.org.uk."""

    url = f"https://polsy.org.uk/stuff/ytrestrict.cgi?agreed=on&ytid={resource}"
    response = requests.get(url, timeout=5)

    if not response.ok:
        return None

    soup = BeautifulSoup(response.text, "html.parser")
    allowed_countries = []

    table = soup.find("table")
    if not table or not isinstance(table, Tag):
        return None

    rows = table.find_all("tr")[1:]  # Skip the header row

    for row in rows:
        if not isinstance(row, Tag):
            continue
        columns = row.find_all("td")
        if len(columns) < 2:
            continue  # Skip invalid rows

        allowed_text = columns[0].get_text(strip=True)

        if allowed_text:
            country_code = allowed_text.split(" - ")[0]
            allowed_countries.append(country_code)

    return allowed_countries


def remove_ansi_codes(text: str) -> str:
    """Removes ANSI escape codes from a string."""

    ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
    return ansi_escape.sub("", text)


def get_first_alphanum_upper(text: str) -> str | None:
    """Returns the first uppercase alphanumeric character in a string."""
    for char in text:
        if char.isalnum():
            return unidecode(char)[0].upper()
    return None


class AppPaths:
    """App data paths."""

    log = Path(_app_dirs.user_data_dir, "usdb_syncer.log")
    db = Path(_app_dirs.user_data_dir, "usdb_syncer.db")
    addons = Path(_app_dirs.user_data_dir, "addons")
    song_list = Path(_app_dirs.user_cache_dir, "available_songs.json")
    profile = Path(_app_dirs.user_cache_dir, "usdb_syncer.prof")
    shared = (_root() / "shared") if IS_SOURCE else None

    @classmethod
    def make_dirs(cls) -> None:
        cls.addons.mkdir(parents=True, exist_ok=True)
        cls.song_list.parent.mkdir(parents=True, exist_ok=True)


class DirectoryCache:
    """Helper to keep track of directories.

    This is to avoid a race condition when two songs requiring the same folder name
    are downloaded concurrently.
    """

    _cache: ClassVar[dict[Path, float]] = {}

    @classmethod
    def insert(cls, path: Path) -> bool:
        """True if path was not in the cache (or the entry had expired)."""
        now = time.time()
        if cls._cache.get(path, 0) + CACHE_LIFETIME < now:
            cls._cache[path] = now
            return True
        return False


def extract_youtube_id(url: str) -> str | None:
    """Extracts the YouTube id from a variety of URLs.

    Partially taken from `https://regexr.com/531i0`.
    """

    pattern = r"""
        (?:https?://)?
        (?:www\.)?
        (?:m\.)?
        (?:
            youtube\.com/
            |
            youtube-nocookie\.com/
            |
            youtu\.be               # no '/' because id may follow immediately
        )
        \S*
        (?:/|%3D|v=|vi=)
        ([0-9a-z_-]{11})            # the actual id
        (?:[%#?&]|$)                # URL may contain additional parameters
        .*
        """
    if match := re.fullmatch(pattern, url, re.VERBOSE | re.IGNORECASE):
        return match.group(1)
    return None


def extract_vimeo_id(url: str) -> str | None:
    """Extracts the Vimeo id from a variety of URLs."""

    pattern = r"""
        (?:https?://)?
        (?:
            www\.
            |
            player\.
        )?
        (?:vimeo\.com/)
        (?:video/)?
        (\d{2,9})                   # the actual id
        (?:[%#?&]|$)                # URL may contain additonal parameters
        .*
        """
    if match := re.search(pattern, url, re.VERBOSE | re.IGNORECASE):
        return match.group(1)
    return None


def read_file_head(
    path: Path, length: int, encoding: str | None = None
) -> list[str] | None:
    """Return the first `length` lines of `path`. If `encoding` is None, try UTF-8 (with
    BOM) first, then cp1252.
    """
    for enc in [encoding] if encoding else ["utf-8-sig", "cp1252"]:
        try:
            with path.open(encoding=enc) as file:
                # strip line break
                return [r[:-1] for r in itertools.islice(file, length)]
        except UnicodeDecodeError:
            pass
    return None


FILENAME_REPLACEMENTS = (('?:"', ""), ("<", "("), (">", ")"), ("/\\|*", "-"))


def sanitize_filename(fname: str) -> str:
    for old, new in FILENAME_REPLACEMENTS:
        for char in old:
            fname = fname.replace(char, new)
    # trailing whitespace is handled inconsistently on OSes, and Windows does not like
    # trailing periods
    return fname.rstrip(" .")


def next_unique_directory(path: Path) -> Path:
    """Ensures directory name is unique by adding a suffix if necessary."""
    out_path = path
    suffix = 0
    while not DirectoryCache.insert(out_path) or out_path.exists():
        suffix += 1
        out_path = path.with_name(f"{path.name} ({suffix})")
    return out_path


def is_name_maybe_with_suffix(text: str, name: str) -> bool:
    "True if `text` is 'name' or 'name (n)' for the provided `name` and some number n."
    if not text.startswith(name):
        return False
    tail = text.removeprefix(name)
    return not tail or re.fullmatch(r" \(\d+\)", tail) is not None


def path_matches_maybe_with_suffix(path: Path, search: Path) -> bool:
    """True if `path` matches `search`, with an optional suffix ` (n)` for some
    number n.
    """
    path = normalize_path(path)
    search = normalize_path(search)
    if path.parent != search.parent:
        return False
    return is_name_maybe_with_suffix(path.name, search.name)


def open_path_or_file(path: Path) -> None:
    logger.debug(f"Opening '{path}' with file explorer.")
    if sys.platform == "win32":
        os.startfile(path)
    elif sys.platform == "linux":
        with LinuxEnvCleaner() as env:
            subprocess.run(["xdg-open", str(path)], check=True, env=env)
    else:
        subprocess.run(["open", str(path)], check=True)


def add_to_system_path(path: str) -> None:
    os.environ["PATH"] = path + os.pathsep + os.environ["PATH"]


def normalize(text: str) -> str:
    """Return the Unicode NFC form of the string."""
    return unicodedata.normalize("NFC", text)


def normalize_path(path: Path) -> Path:
    return Path(*(normalize(p) for p in path.parts))


def compare_unicode_paths(lhs: Path, rhs: Path) -> bool:
    """Checks two paths for equality, taking into account implicit Unicode normalization
    on macOS.
    """
    if sys.platform == "darwin":
        return normalize(str(lhs)) == normalize(str(rhs))
    return lhs == rhs


def resource_file_ending(name: str) -> str:
    """Return the suffix or name, including " [BG]" and " [CO]"."""
    regex = re.compile(r".+?((?: \[(?:CO|BG)\])?\.[^.]+)")
    if match := regex.fullmatch(name):
        return match.group(1)
    return ""


def get_mtime(path: Path) -> int:
    """Helper for mtime in microseconds so it can be stored in db losslessly."""
    return int(path.stat().st_mtime * 1_000_000)


@functools.cache
def format_timestamp(micros: int) -> str:
    return datetime.datetime.fromtimestamp(micros / 1_000_000).strftime(
        "%Y-%m-%d %H:%M:%S"
    )


def get_latest_version() -> str | None:
    response = requests.get(constants.GITHUB_API_LATEST, timeout=5)
    if response.status_code == 200:
        return response.json()["tag_name"]
    return None


def newer_version_available() -> str | None:
    if latest_version := get_latest_version():
        if version.parse(usdb_syncer.__version__) < version.parse(latest_version):
            logger.warning(
                f"USDB Syncer {latest_version} is available! "
                f"(You have {usdb_syncer.__version__}). Please download the latest "
                f"release from {constants.GITHUB_DL_LATEST}."
            )
            return latest_version
        logger.info(f"You are running the latest Syncer version {latest_version}.")
        return None

    logger.info("Could not determine the latest version.")
    return None


def start_process_detached(command: list[str]) -> subprocess.Popen:
    """Start a process in a fully detached mode, cross-platform."""
    # We are not using a context manager here so that the app is launched
    # without blocking the syncer.
    flags = 0
    if sys.platform == "win32":
        flags = subprocess.DETACHED_PROCESS | subprocess.CREATE_NEW_PROCESS_GROUP
    with LinuxEnvCleaner() as env:
        return subprocess.Popen(command, creationflags=flags, close_fds=True, env=env)


class LinuxEnvCleaner:
    """Context manager to clean an environment. Specifically, it removes paths starting
    with /tmp/_MEI and some specific Qt-related paths from the environment.

    This is needed when running applications bundled with PyInstaller, as it links some
    libraries that may interfere with dynamically loaded libraries.

    This class is only useful on Linux systems, where it modifies the environment to
    avoid conflicts.
    On non-Linux platforms, it is a no-op and does not alter the environment.
    """

    def __init__(self) -> None:
        self.env = os.environ.copy()
        self.modified_env = self.env.copy()
        self.old_values: dict[str, str] = {}

    def __enter__(self) -> dict[str, str]:
        if sys.platform == "linux":
            for key, value in self.env.items():
                if key == "LD_LIBRARY_PATH":
                    new_value = ":".join(
                        path
                        for path in (value.split(":") if value else [])
                        if not path.startswith("/tmp/_MEI")  # noqa: S108
                    )
                    self.old_values[key] = value
                    self.modified_env[key] = new_value
                if key in (
                    "QT_PLUGIN_PATH",
                    "QT_QPA_PLATFORM_PLUGIN_PATH",
                    "QT_DEBUG_PLUGINS",
                ):
                    self.old_values[key] = value
                    self.modified_env.pop(key)
            os.environ.clear()
            os.environ.update(self.modified_env)
        return self.modified_env.copy()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        if sys.platform == "linux":
            current_env = os.environ.copy()

            for key, value in current_env.items():
                if key not in self.env:
                    self.modified_env[key] = value
            for key, value in self.old_values.items():
                self.modified_env[key] = value

            os.environ.clear()
            os.environ.update(self.modified_env)



"""Functions for downloading and processing media."""

import io
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import assert_never

import filetype
import requests
import yt_dlp
from PIL import Image, ImageEnhance, ImageOps
from PIL.Image import Resampling

from usdb_syncer import utils
from usdb_syncer.constants import YtErrorMsg
from usdb_syncer.download_options import AudioOptions, VideoOptions
from usdb_syncer.logger import Logger, song_logger
from usdb_syncer.meta_tags import ImageMetaTags
from usdb_syncer.postprocessing import normalize_audio
from usdb_syncer.settings import (
    AudioNormalization,
    Browser,
    CoverMaxSize,
    YtdlpRateLimit,
)
from usdb_syncer.usdb_scraper import SongDetails
from usdb_syncer.utils import video_url_from_resource

IMAGE_DOWNLOAD_HEADERS = {
    "User-Agent": (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
        "(KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36"
    )
}

YtdlOptions = dict[str, str | bool | tuple | list | int]


class ResourceDLError(Enum):
    """Errors that can occur when downloading a resource."""

    RESOURCE_INVALID = "resource invalid"
    RESOURCE_UNSUPPORTED = "resource unsupported"
    RESOURCE_GEO_RESTRICTED = "resource geo-restricted"
    RESOURCE_UNAVAILABLE = "resource unavailable"
    RESOURCE_PARSE_ERROR = "resource parse error"
    RESOURCE_DL_FAILED = "resource download failed"
    RESOURCE_FORBIDDEN = "resource forbidden"


@dataclass
class ResourceDLResult:
    """The result of a download operation."""

    extension: str | None = None
    error: ResourceDLError | None = None


class ImageKind(Enum):
    """Types of images used for songs."""

    COVER = "CO"
    BACKGROUND = "BG"

    def __str__(self) -> str:
        match self:
            case ImageKind.COVER:
                return "cover"
            case ImageKind.BACKGROUND:
                return "background"
            case _ as unreachable:
                assert_never(unreachable)


def download_audio(
    resource: str,
    options: AudioOptions,
    browser: Browser,
    path_stem: Path,
    logger: Logger,
) -> ResourceDLResult:
    """Download audio from resource to path and process it according to options.

    Parameters:
        resource: URL or YouTube id
        options: parameters for downloading and processing
        browser: browser to use cookies from
        path_stem: the target on the file system *without* an extension
        logger: logger

    Returns:
        DownloadResult with the extension of the (postprocessed, possibly normalized)
        audio file, if successful
    """
    ydl_opts = _ytdl_options(
        options.ytdl_format(), browser, path_stem, options.rate_limit
    )
    if options.normalization in {
        AudioNormalization.DISABLE,
        AudioNormalization.REPLAYGAIN,
    }:
        # DISABLE or REPLAYGAIN normalization will not re-encode the audio file to
        # target format, so we have to add a postprocessor to get it
        postprocessor = {
            "key": "FFmpegExtractAudio",
            "preferredquality": options.bitrate.ytdl_format(),
            "preferredcodec": options.format.ytdl_codec(),
        }
        ydl_opts["postprocessors"] = [postprocessor]

    dl_result = _download_resource(resource, ydl_opts, logger)
    if not dl_result.extension:
        return dl_result
    if options.normalization is not AudioNormalization.DISABLE:
        normalize_audio(options, path_stem, dl_result.extension, logger)

    # either way, the resulting file is in target format, so we have to correct the
    # extension before returning dl_result
    dl_result.extension = options.format.value
    return dl_result


def download_video(
    resource: str,
    options: VideoOptions,
    browser: Browser,
    path_stem: Path,
    logger: Logger,
) -> ResourceDLResult:
    """Download video from resource to path and process it according to options.

    Parameters:
        resource: URL or YouTube id
        options: parameters for downloading and processing
        browser: browser to use cookies from
        path_stem: the target on the file system *without* an extension

    Returns:
        DownloadResult with the extension of the downloaded file if successful
    """
    ydl_opts = _ytdl_options(
        options.ytdl_format(), browser, path_stem, options.rate_limit
    )
    return _download_resource(resource, ydl_opts, logger)


def _ytdl_options(
    format_: str, browser: Browser, target_stem: Path, ratelimit: YtdlpRateLimit
) -> YtdlOptions:
    options: YtdlOptions = {
        "format": format_,
        "outtmpl": f"{target_stem}.%(ext)s",
        "keepvideo": False,
        "verbose": False,
        # suppresses download of playlists, channels and search results
        "playlistend": 0,
        "overwrites": True,
    }
    if ratelimit.value is not None:
        options["ratelimit"] = ratelimit.value
    if browser:
        options["cookiesfrombrowser"] = (browser.value, None, None, None)
    return options


def _download_resource(
    resource: str, options: YtdlOptions, logger: Logger
) -> ResourceDLResult:
    if (url := video_url_from_resource(resource)) is None:
        return ResourceDLResult(error=ResourceDLError.RESOURCE_INVALID)

    options_without_cookies = options.copy()
    options_without_cookies.pop("cookiesfrombrowser", None)

    with yt_dlp.YoutubeDL(options_without_cookies) as ydl:
        try:
            filename = ydl.prepare_filename(ydl.extract_info(url))
            ext = Path(filename).suffix[1:]
            return ResourceDLResult(extension=ext)
        except yt_dlp.utils.UnsupportedError:
            return ResourceDLResult(error=ResourceDLError.RESOURCE_UNSUPPORTED)
        except yt_dlp.utils.YoutubeDLError as e:
            error_message = utils.remove_ansi_codes(str(e))
            logger.debug(f"Failed to download '{url}': {error_message}")
            if YtErrorMsg.YT_AGE_RESTRICTED in error_message:
                dl_result = _retry_with_cookies(url, options, logger)
                return ResourceDLResult(extension=dl_result.extension)
            if YtErrorMsg.YT_GEO_RESTRICTED in error_message:
                _handle_geo_restriction(url, resource, logger)
                return ResourceDLResult(error=ResourceDLError.RESOURCE_GEO_RESTRICTED)
            if YtErrorMsg.YT_UNAVAILABLE in error_message:
                _handle_unavailable(url, logger)
                return ResourceDLResult(error=ResourceDLError.RESOURCE_UNAVAILABLE)
            if YtErrorMsg.YT_PARSE_ERROR in error_message:
                _handle_parse_error(url, logger)
                return ResourceDLResult(error=ResourceDLError.RESOURCE_PARSE_ERROR)
            if YtErrorMsg.YT_FORBIDDEN in error_message:
                _handle_forbidden(url, logger)
                return ResourceDLResult(error=ResourceDLError.RESOURCE_FORBIDDEN)
            raise


def _retry_with_cookies(
    url: str, options: YtdlOptions, logger: Logger
) -> ResourceDLResult:
    logger.warning("Age-restricted resource. Retrying with cookies ...")
    with yt_dlp.YoutubeDL(options) as ydl:
        try:
            filename = ydl.prepare_filename(ydl.extract_info(url))
            ext = Path(filename).suffix[1:]
            return ResourceDLResult(extension=ext)
        except yt_dlp.utils.YoutubeDLError as re:
            msg = f"Retry failed: {utils.remove_ansi_codes(str(re))}"
            logger.error(msg)  # noqa: TRY400
            raise


def _handle_geo_restriction(url: str, resource: str, logger: Logger) -> None:
    logger.warning("Geo-restricted resource. You can retry after connecting to a VPN.")
    if "youtube" in url and (
        allowed_countries := utils.get_allowed_countries(resource)
    ):
        logger.info(
            "Countries where the resource is available: " + ", ".join(allowed_countries)
        )


def _handle_unavailable(url: str, logger: Logger) -> None:
    logger.warning(
        f"Resource '{url}' is no longer available. Please support the community, "
        "find a suitable replacement resource and comment it on USDB."
    )


def _handle_parse_error(url: str, logger: Logger) -> None:
    logger.warning(f"Failed to parse XML for resource '{url}'.")


def _handle_forbidden(url: str, logger: Logger) -> None:
    logger.warning(
        f"Failed to download resource '{url}'. Your IP/account might have been blocked."
    )


def download_image(url: str, logger: Logger) -> bytes | None:
    try:
        reply = requests.get(
            url, allow_redirects=True, headers=IMAGE_DOWNLOAD_HEADERS, timeout=60
        )
    except requests.exceptions.SSLError:
        logger.exception(
            f"Failed to retrieve {url}. The SSL certificate could not be verified."
        )
        return None
    except requests.RequestException:
        logger.exception(
            f"Failed to retrieve {url}. The URL might be invalid, the server may be "
            "down or your internet connection is currently unavailable."
        )
        return None
    if reply.status_code in range(100, 299):
        # 1xx informational response, 2xx success
        return reply.content
    if reply.status_code in range(300, 399):
        # 3xx redirection
        logger.debug(
            f"'{url}' redirects to '{reply.headers['Location']}'. "
            "Please adapt metatags."
        )
        return reply.content
    if reply.status_code in range(400, 499):
        logger.error(
            f"Client error {reply.status_code}. Failed to download {reply.url}"
        )
    elif reply.status_code in range(500, 599):
        logger.error(
            f"Server error {reply.status_code}. Failed to download {reply.url}"
        )
    return None


def download_and_process_image(
    url: str,
    *,
    target_stem: Path,
    meta_tags: ImageMetaTags | None,
    details: SongDetails,
    kind: ImageKind,
    max_width: CoverMaxSize | None,
    process: bool = True,
) -> Path | None:
    logger = song_logger(details.song_id)
    if not (img_bytes := download_image(url, logger)):
        logger.error(f"#{str(kind).upper()}: file does not exist at url: {url}")
        return None

    if not filetype.is_image(img_bytes):
        logger.error(f"#{str(kind).upper()}: file at {url} is not an image")
        return None

    path = target_stem.with_name(f"{target_stem.name} [{kind.value}].jpg")
    with Image.open(io.BytesIO(img_bytes)).convert("RGB") as img:
        img.save(path)

    if process:
        _process_image(meta_tags, kind, max_width, path)
    return path


def _rotate(image: Image.Image, meta_tags: ImageMetaTags) -> Image.Image:
    if rotate := meta_tags.rotate:
        return image.rotate(rotate, resample=Resampling.BICUBIC, expand=True)
    return image


def _crop(image: Image.Image, meta_tags: ImageMetaTags) -> Image.Image:
    if crop := meta_tags.crop:
        return image.crop((crop.left, crop.upper, crop.right, crop.lower))
    return image


def _resize(image: Image.Image, meta_tags: ImageMetaTags) -> Image.Image:
    if resize := meta_tags.resize:
        return image.resize((resize.width, resize.height), resample=Resampling.LANCZOS)
    return image


def _adjust_contrast(image: Image.Image, meta_tags: ImageMetaTags) -> Image.Image:
    if meta_tags.contrast == "auto":
        return ImageOps.autocontrast(image, cutoff=5)
    if meta_tags.contrast:
        return ImageEnhance.Contrast(image).enhance(meta_tags.contrast)
    return image


def _process_image(
    meta_tags: ImageMetaTags | None,
    kind: ImageKind,
    max_width: CoverMaxSize | None,
    path: Path,
) -> None:
    processed = False
    with Image.open(path).convert("RGB") as image:
        if meta_tags and meta_tags.image_processing():
            processed = True
            match kind:
                case ImageKind.COVER:
                    operations = [_rotate, _crop, _resize, _adjust_contrast]
                case ImageKind.BACKGROUND:
                    operations = [_resize, _crop]
                case _ as unreachable:
                    assert_never(unreachable)

            for operation in operations:
                image = operation(image, meta_tags)

        if (
            max_width
            and max_width != CoverMaxSize.DISABLE
            and max_width.value < image.width
        ):
            processed = True
            height = round(image.height * max_width.value / image.width)
            image = image.resize((max_width.value, height), resample=Resampling.LANCZOS)

        if processed:
            image.save(path, "jpeg", quality=100, subsampling=0)



"""Controller for the filter tree."""

from __future__ import annotations

import copy
from collections.abc import Callable
from typing import TYPE_CHECKING

from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import QModelIndex, Qt

from usdb_syncer import db, events
from usdb_syncer.gui.gui_utils import keyboard_modifiers

from .item import Filter, SavedSearch
from .model import TreeModel, TreeProxyModel

if TYPE_CHECKING:
    from usdb_syncer.gui.mw import MainWindow


class FilterTree:
    """Controller for the filter tree."""

    _search = db.SearchBuilder()

    def __init__(self, mw: MainWindow) -> None:
        self.mw = mw
        self.view = mw.search_view
        self._model = TreeModel(mw)
        self._proxy_model = TreeProxyModel(self.view, self._model)
        self.view.setHeaderHidden(True)
        self.view.setModel(self._proxy_model)
        self.view.clicked.connect(self._on_click)
        self.view.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.view.customContextMenuRequested.connect(self._context_menu)
        self._model.dataChanged.connect(self._on_data_changed)
        mw.line_edit_search_filters.textChanged.connect(self._proxy_model.set_filter)
        self._setup_actions()
        events.SearchOrderChanged.subscribe(self._on_search_order_changed)
        events.TextFilterChanged.subscribe(self._on_text_filter_changed)
        events.SavedSearchRestored.subscribe(self._restore_saved_search)

    def _setup_actions(self) -> None:
        self.mw.action_add_saved_search.triggered.connect(self._add_saved_search)
        self.mw.action_update_saved_search.triggered.connect(self._update_saved_search)
        self.mw.action_delete_saved_search.triggered.connect(self._delete_saved_search)
        self.mw.action_set_saved_search_default.triggered.connect(
            self._set_search_is_default
        )
        self.mw.action_set_saved_search_subscribed.triggered.connect(
            self._set_search_subscribed
        )

    def populate(self) -> None:
        self._model.populate()
        self.view.expand(
            self._proxy_model.mapFromSource(
                self._model.index_for_item(self._model.root.children[0])
            )
        )

    def _restore_saved_search(self, event: events.SavedSearchRestored) -> None:
        for filt in self._model.root.children:
            for changed in filt.set_checked_children(event.search):
                self._model.emit_item_changed(changed)

    def _on_click(self, index: QModelIndex) -> None:
        if not (
            item := self._model.item_for_index(self._proxy_model.mapToSource(index))
        ):
            return
        if isinstance(item.data, SavedSearch):
            events.SavedSearchRestored(item.data.search).post()
        else:
            for changed in item.toggle_checked(keyboard_modifiers().ctrl):
                self._model.emit_item_changed(changed)

    def connect_filter_changed(self, func: Callable[[], None]) -> None:
        self._model.dataChanged.connect(func)

    def _on_data_changed(self) -> None:
        self._search = db.SearchBuilder(
            order=self._search.order,
            descending=self._search.descending,
            text=self._search.text,
        )
        for filt in self._model.root.children:
            filt.build_search(self._search)
        events.TreeFilterChanged(self._search).post()

    def _context_menu(self, _pos: QtCore.QPoint) -> None:
        item = self._model.item_for_index(
            self._proxy_model.mapToSource(self.view.currentIndex())
        )
        if item and item.data == Filter.SAVED:
            actions = [self.mw.action_add_saved_search]
        elif item and isinstance(item.data, SavedSearch):
            self.mw.action_set_saved_search_default.setChecked(item.data.is_default)
            self.mw.action_set_saved_search_subscribed.setChecked(item.data.subscribed)
            actions = [
                self.mw.action_update_saved_search,
                self.mw.action_delete_saved_search,
                self.mw.action_set_saved_search_default,
                self.mw.action_set_saved_search_subscribed,
            ]
        else:
            return
        menu = QtWidgets.QMenu()
        menu.addActions(actions)
        menu.exec(QtGui.QCursor.pos())

    def _delete_saved_search(self) -> None:
        self._model.delete_saved_search(
            self._proxy_model.mapToSource(self.view.currentIndex())
        )

    def _get_current_saved_search(self) -> SavedSearch | None:
        item = self._model.item_for_index(
            self._proxy_model.mapToSource(self.view.currentIndex())
        )
        if item and isinstance(item.data, SavedSearch):
            return item.data
        return None

    def _update_saved_search(self) -> None:
        if search := self._get_current_saved_search():
            search.search = copy.deepcopy(self._search)
            with db.transaction():
                search.update()

    def _add_saved_search(self) -> None:
        index = self._proxy_model.mapFromSource(
            self._model.insert_saved_search(SavedSearch("My search", self._search))
        )
        self.view.setCurrentIndex(index)
        self.view.edit(index)

    def _set_search_is_default(self, default: bool) -> None:
        if search := self._get_current_saved_search():
            if default:
                for item in self._model.root.children[0].children:
                    if isinstance(item.data, SavedSearch) and item.data.is_default:
                        item.data.is_default = False
            search.is_default = default
            with db.transaction():
                search.update()

    def _set_search_subscribed(self, subscribe: bool) -> None:
        if search := self._get_current_saved_search():
            search.subscribed = subscribe
            with db.transaction():
                search.update()

    def _on_search_order_changed(self, event: events.SearchOrderChanged) -> None:
        self._search.descending = event.descending
        self._search.order = event.order

    def _on_text_filter_changed(self, event: events.TextFilterChanged) -> None:
        self._search.text = event.search
